import os
import pathlib
import random
import sys
import time

import mpi4py.util.sync as sync
from mpi4py import MPI

try:
    import mpiunittest as unittest
except ImportError:
    sys.path.append(os.fspath(pathlib.Path(__file__).resolve().parent))
    import mpiunittest as unittest

# ---


def random_sleep(max_sleep=0.01):
    time.sleep(max_sleep * random.random())


# ---


class BaseTestSequential:
    #
    COMM = MPI.COMM_NULL

    def testWith(self):
        comm = self.COMM
        for _ in range(3):
            counter = sync.Counter(comm=comm)
            with sync.Sequential(comm):
                value = next(counter)
            counter.free()
            self.assertEqual(
                comm.allgather(value),
                list(range(comm.size)),
            )

    def testBeginEnd(self):
        comm = self.COMM
        seq = sync.Sequential(comm)
        for _ in range(3):
            counter = sync.Counter(comm=comm)
            seq.begin()
            value = next(counter)
            seq.end()
            counter.free()
            self.assertEqual(
                comm.allgather(value),
                list(range(comm.size)),
            )


class TestSequentialSelf(BaseTestSequential, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


class TestSequentialWorld(BaseTestSequential, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


# ---


class BaseTestCounter:
    #
    COMM = MPI.COMM_NULL

    def testIter(self):
        comm = self.COMM
        size = comm.Get_size()
        counter = sync.Counter(comm=comm)
        for value in counter:
            random_sleep()
            if value >= size - 1:
                break
        counter.free()

    def testNext(self):
        comm = self.COMM
        size = comm.Get_size()
        counter = sync.Counter(comm=comm)
        while True:
            value = next(counter)
            random_sleep()
            if value >= size - 1:
                break
        counter.free()

    def execute(self, counter, its, condition=True):
        values = []
        if condition:
            for _ in range(its):
                value = counter.next()
                values.append(value)
        return sorted(self.COMM.allreduce(values))

    def testDefault(self):
        comm = self.COMM
        size = comm.Get_size()
        counter = sync.Counter(comm=comm)
        values = self.execute(counter, 5)
        counter.free()
        self.assertEqual(values, list(range(5 * size)))

    def testStart(self):
        comm = self.COMM
        size = comm.Get_size()
        counter = sync.Counter(start=7, comm=comm)
        values = self.execute(counter, 5)
        counter.free()
        self.assertEqual(values, list(range(7, 7 + 5 * size)))

    def testStep(self):
        comm = self.COMM
        size = comm.Get_size()
        counter = sync.Counter(step=2, comm=comm)
        values = self.execute(counter, 5)
        counter.free()
        self.assertEqual(values, list(range(0, 2 * 5 * size, 2)))

    def testTypechar(self):
        comm = self.COMM
        size = comm.Get_size()
        for typechar in "hHiIlLqQ" + "fd":
            counter = sync.Counter(typecode=typechar, comm=comm)
            values = self.execute(counter, 3)
            counter.free()
            self.assertEqual(values, list(range(3 * size)))

    def testRoot(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        for root in range(size):
            counter = sync.Counter(comm=comm, root=root)
            values = self.execute(counter, 5, rank != root)
            counter.free()
            self.assertEqual(values, list(range(5 * (size - 1))))

    def testFree(self):
        comm = self.COMM
        counter = sync.Counter(comm=comm)
        for _ in range(5):
            counter.free()
        self.assertRaises(RuntimeError, counter.next)


class TestCounterSelf(BaseTestCounter, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


class TestCounterWorld(BaseTestCounter, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


# ---


class BaseTestMutexBasic:
    #
    COMM = MPI.COMM_NULL

    def setUp(self):
        self.mutex = sync.Mutex(comm=self.COMM)

    def tearDown(self):
        self.mutex.free()

    def testExclusion(self):
        comm = self.COMM
        mutex = self.mutex
        counter = sync.Counter(comm=comm)
        values = []
        comm.Barrier()
        mutex.acquire()
        for _ in range(10):
            values.append(next(counter))
        mutex.release()
        counter.free()
        first, last = values[0], values[-1]
        self.assertEqual(first % 10, 0)
        self.assertEqual(last - first + 1, 10)
        self.assertEqual(values, list(range(first, last + 1)))

    def testFairness(self):
        comm = self.COMM
        mutex = self.mutex
        number = 0
        counter = sync.Counter(comm=comm)
        while next(counter) < comm.Get_size() * 5:
            mutex.acquire()
            number += 1
            mutex.release()
            comm.Barrier()
        counter.free()
        self.assertEqual(number, 5)

    def testWith(self):
        def test_with():
            mutex = self.mutex
            self.assertFalse(mutex.locked())
            self.assertRaises(RuntimeError, mutex.release)
            with mutex:
                self.assertTrue(mutex.locked())
                self.assertRaises(RuntimeError, mutex.acquire)
            self.assertFalse(mutex.locked())
            self.assertRaises(RuntimeError, mutex.release)

        for _ in range(5):
            self.COMM.Barrier()
            test_with()
        for _ in range(5):
            random_sleep()
            test_with()

    def testAcquireRelease(self):
        def test_acquire_release():
            mutex = self.mutex
            self.assertFalse(mutex.locked())
            self.assertRaises(RuntimeError, mutex.release)
            locked = mutex.acquire()
            self.assertTrue(locked)
            self.assertTrue(mutex.locked())
            self.assertRaises(RuntimeError, mutex.acquire)
            mutex.release()
            self.assertFalse(mutex.locked())
            self.assertRaises(RuntimeError, mutex.release)

        for _ in range(5):
            self.COMM.Barrier()
            test_acquire_release()
        for _ in range(5):
            random_sleep()
            test_acquire_release()

    def testAcquireNonblocking(self):
        def test_acquire_nonblocking():
            comm = self.COMM
            mutex = self.mutex
            self.assertFalse(mutex.locked())
            comm.Barrier()
            locked = mutex.acquire(blocking=False)
            comm.Barrier()
            self.assertEqual(mutex.locked(), locked)
            if locked:
                mutex.release()
            self.assertFalse(mutex.locked())
            states = comm.allgather(locked)
            self.assertEqual(states.count(True), 1)
            comm.Barrier()
            while not mutex.acquire(blocking=False):
                pass
            mutex.release()
            comm.Barrier()

        for _ in range(5):
            self.COMM.Barrier()
            test_acquire_nonblocking()
        for _ in range(5):
            random_sleep()
            test_acquire_nonblocking()

    def testAcquireFree(self):
        mutex = self.mutex
        mutex.acquire()
        for _ in range(5):
            mutex.free()
        self.assertRaises(RuntimeError, mutex.acquire)
        self.assertRaises(RuntimeError, mutex.release)
        self.assertRaises(RuntimeError, mutex.locked)

    def testFree(self):
        mutex = self.mutex
        for _ in range(5):
            mutex.free()
        self.assertRaises(RuntimeError, mutex.acquire)
        self.assertRaises(RuntimeError, mutex.release)
        self.assertRaises(RuntimeError, mutex.locked)


class TestMutexBasicSelf(BaseTestMutexBasic, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


@unittest.skipMPI("msmpi")
class TestMutexBasicWorld(BaseTestMutexBasic, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD

    @unittest.skipMPI("msmpi")
    def testExclusion(self):
        super().testExclusion()


# ---


class BaseTestMutexRecursive:
    #
    COMM = MPI.COMM_NULL

    def setUp(self):
        self.mutex = sync.Mutex(recursive=True, comm=self.COMM)

    def tearDown(self):
        self.mutex.free()

    def testWith(self):
        def test_with():
            mutex = self.mutex
            self.assertEqual(mutex.count(), 0)
            self.assertRaises(RuntimeError, mutex.release)
            with mutex:
                self.assertEqual(mutex.count(), 1)
                with mutex:
                    self.assertEqual(mutex.count(), 2)
                    with mutex:
                        self.assertEqual(mutex.count(), 3)
                    self.assertEqual(mutex.count(), 2)
                self.assertEqual(mutex.count(), 1)
            self.assertEqual(mutex.count(), 0)
            self.assertRaises(RuntimeError, mutex.release)

        for _ in range(5):
            self.COMM.Barrier()
            test_with()
        for _ in range(5):
            random_sleep()
            test_with()

    def testAcquireRelease(self):
        def test_acquire_release():
            mutex = self.mutex
            self.assertFalse(mutex.locked())
            self.assertEqual(mutex.count(), 0)
            self.assertRaises(RuntimeError, mutex.release)
            mutex.acquire()
            self.assertTrue(mutex.locked())
            self.assertEqual(mutex.count(), 1)
            mutex.acquire()
            self.assertTrue(mutex.locked())
            self.assertEqual(mutex.count(), 2)
            mutex.acquire()
            self.assertTrue(mutex.locked())
            self.assertEqual(mutex.count(), 3)
            mutex.release()
            self.assertTrue(mutex.locked())
            self.assertEqual(mutex.count(), 2)
            mutex.release()
            self.assertTrue(mutex.locked())
            self.assertEqual(mutex.count(), 1)
            mutex.release()
            self.assertFalse(mutex.locked())
            self.assertEqual(mutex.count(), 0)
            self.assertRaises(RuntimeError, mutex.release)

        for _ in range(5):
            self.COMM.Barrier()
            test_acquire_release()
        for _ in range(5):
            random_sleep()
            test_acquire_release()

    def testAcquireNonblocking(self):
        def test_acquire_nonblocking():
            comm = self.COMM
            mutex = self.mutex
            self.assertEqual(mutex.count(), 0)
            comm.Barrier()
            locked = mutex.acquire(blocking=False)
            self.assertEqual(mutex.locked(), locked)
            comm.Barrier()
            self.assertEqual(mutex.count(), int(locked))
            if locked:
                self.assertEqual(mutex.count(), 1)
                flag = mutex.acquire(blocking=False)
                self.assertTrue(flag)
                self.assertEqual(mutex.count(), 2)
                flag = mutex.acquire(blocking=True)
                self.assertTrue(flag)
                self.assertEqual(mutex.count(), 3)
                mutex.release()
                self.assertEqual(mutex.count(), 2)
                mutex.release()
                self.assertEqual(mutex.count(), 1)
                mutex.release()
            comm.Barrier()
            self.assertFalse(mutex.locked())
            self.assertEqual(mutex.count(), 0)
            states = comm.allgather(locked)
            self.assertEqual(states.count(True), 1)
            comm.Barrier()
            while not mutex.acquire(blocking=False):
                pass
            mutex.release()
            comm.Barrier()

        for _ in range(5):
            self.COMM.Barrier()
            test_acquire_nonblocking()
        for _ in range(5):
            random_sleep()
            test_acquire_nonblocking()

    def testAcquireFree(self):
        mutex = self.mutex
        mutex.acquire()
        mutex.acquire()
        mutex.acquire()
        for _ in range(5):
            mutex.free()
        self.assertRaises(RuntimeError, mutex.acquire)
        self.assertRaises(RuntimeError, mutex.release)
        self.assertRaises(RuntimeError, mutex.count)

    def testFree(self):
        mutex = self.mutex
        for _ in range(5):
            mutex.free()
        self.assertRaises(RuntimeError, mutex.acquire)
        self.assertRaises(RuntimeError, mutex.release)
        self.assertRaises(RuntimeError, mutex.count)


class TestMutexRecursiveSelf(BaseTestMutexRecursive, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


@unittest.skipMPI("msmpi")
class TestMutexRecursiveWorld(BaseTestMutexRecursive, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


# ---


class BaseTestCondition:
    #
    COMM = MPI.COMM_NULL

    def setUp(self):
        self.mutex = None
        self.condition = sync.Condition(comm=self.COMM)

    def tearDown(self):
        self.condition.free()

    def testWaitNotify(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        cv = self.condition
        if rank == 0:
            with cv:
                num = cv.notify()
            self.assertEqual(num, 0)
            comm.Barrier()
            while num < size - 1:
                with cv:
                    num += cv.notify()
                    random_sleep()
            with cv:
                num = cv.notify()
            self.assertEqual(num, 0)
        else:
            comm.Barrier()
            with cv:
                random_sleep()
                cv.wait()
        self.assertRaises(RuntimeError, cv.wait)
        self.assertRaises(RuntimeError, cv.notify)

    def testWaitForNotify(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        cv = self.condition
        if rank == 0:
            with cv:
                num = cv.notify()
            self.assertEqual(num, 0)
            comm.Barrier()
            reqs1 = [comm.isend(00, dest) for dest in range(1, size)]
            reqs2 = [comm.isend(42, dest) for dest in range(1, size)]
            while num < size - 1:
                MPI.Request.Testall(reqs1)
                with cv:
                    num += cv.notify()
                    random_sleep()
            self.assertEqual(num, size - 1)
            MPI.Request.Waitall(reqs2)
            with cv:
                num = cv.notify()
            self.assertEqual(num, 0)
        else:
            comm.Barrier()
            with cv:
                random_sleep()
                result = cv.wait_for(lambda: comm.recv())
            self.assertEqual(result, 42)
        self.assertRaises(RuntimeError, cv.wait_for, lambda: False)
        self.assertRaises(RuntimeError, cv.notify)

    def testWaitNotifyAll(self):
        comm = self.COMM
        size = comm.Get_size()
        rank = comm.Get_rank()
        cv = self.condition
        if rank == 0:
            with cv:
                num = cv.notify_all()
            self.assertEqual(num, 0)
            comm.Barrier()
            while num < size - 2:
                with cv:
                    num += cv.notify_all()
                    random_sleep()
            self.assertEqual(num, max(0, size - 2))
            with cv:
                num = cv.notify()
            self.assertEqual(num, 0)
        elif rank == 1:
            comm.Barrier()
        else:
            comm.Barrier()
            with cv:
                random_sleep()
                cv.wait()
        self.assertRaises(RuntimeError, cv.wait)
        self.assertRaises(RuntimeError, cv.notify_all)

    def testAcquireFree(self):
        cv = self.condition
        cv.acquire()
        for _ in range(5):
            cv.free()

    def testFree(self):
        cv = self.condition
        for _ in range(5):
            cv.free()
        self.assertRaises(RuntimeError, cv.acquire)
        self.assertRaises(RuntimeError, cv.release)
        self.assertRaises(RuntimeError, cv.wait)
        self.assertRaises(RuntimeError, cv.wait_for, lambda: False)
        self.assertRaises(RuntimeError, cv.notify)
        self.assertRaises(RuntimeError, cv.notify_all)


class TestConditionSelf(BaseTestCondition, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


@unittest.skipMPI("msmpi")
class TestConditionWorld(BaseTestCondition, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


# ---


class BaseTestConditionMutex(BaseTestCondition):
    #
    COMM = MPI.COMM_NULL

    def setUp(self):
        comm = self.COMM
        self.mutex = sync.Mutex(comm=comm)
        self.condition = sync.Condition(self.mutex)

    def tearDown(self):
        self.mutex.free()
        self.condition.free()


class TestConditionMutexSelf(BaseTestConditionMutex, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


@unittest.skipMPI("msmpi")
class TestConditionMutexWorld(BaseTestConditionMutex, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


# ---


class BaseTestSemaphore:
    #
    COMM = MPI.COMM_NULL

    def setUp(self):
        comm = self.COMM
        value = max(1, comm.Get_size() - 1)
        self.semaphore = sync.Semaphore(value, bounded=False, comm=comm)

    def tearDown(self):
        self.semaphore.free()

    def testValue(self):
        sem = self.semaphore
        self.assertRaises(ValueError, sem.release, 0)
        self.assertRaises(ValueError, sem.release, -1)
        self.assertRaises(ValueError, sync.Semaphore, -1)

    def testBounded(self):
        sem = self.semaphore
        comm = self.COMM
        count = max(1, comm.size - 1)
        sem._bounded = False
        if comm.size > 1:
            if comm.rank == 0:
                sem.release()
        comm.Barrier()
        self.assertEqual(sem._counter.next(0), comm.size)
        comm.Barrier()
        if comm.size > 1:
            if comm.rank == 0:
                sem.acquire()
        comm.Barrier()
        self.assertEqual(sem._counter.next(0), count)
        sem._bounded = True
        self.assertRaises(ValueError, sem.release)
        comm.Barrier()
        sem.acquire()
        sem.release()
        comm.Barrier()
        self.assertEqual(sem._counter.next(0), count)

    def testWith(self):
        def test_with():
            sem = self.semaphore
            with sem:
                pass

        for _ in range(5):
            self.COMM.Barrier()
            test_with()
        for _ in range(5):
            random_sleep()
            test_with()

    def testAcquireRelease(self):
        def test_acquire_release():
            sem = self.semaphore
            locked = sem.acquire()
            sem.release()
            self.assertTrue(locked)

        for _ in range(5):
            self.COMM.Barrier()
            test_acquire_release()
        for _ in range(5):
            random_sleep()
            test_acquire_release()

    def testAcquireNonblocking(self):
        def test_acquire_nonblocking():
            sem = self.semaphore
            comm = self.COMM
            count = max(1, comm.size - 1)
            comm.Barrier()
            locked = sem.acquire(blocking=False)
            comm.Barrier()
            self.assertEqual(sem._counter.next(0), 0)
            comm.Barrier()
            if locked:
                sem.release()
            comm.Barrier()
            states = comm.allgather(locked)
            self.assertEqual(states.count(True), count)
            self.assertEqual(sem._counter.next(0), count)
            comm.Barrier()
            while not sem.acquire(blocking=False):
                random_sleep()
            sem.release()
            comm.Barrier()
            self.assertEqual(sem._counter.next(0), count)

        for _ in range(5):
            self.COMM.Barrier()
            test_acquire_nonblocking()
        for _ in range(5):
            random_sleep()
            test_acquire_nonblocking()

    def testAcquireFree(self):
        comm = self.COMM
        sem = self.semaphore
        if comm.rank > 0:
            sem.acquire()
        for _ in range(5):
            sem.free()

    def testFree(self):
        sem = self.semaphore
        for _ in range(5):
            sem.free()
        self.assertRaises(RuntimeError, sem.acquire)
        self.assertRaises(RuntimeError, sem.release)


class TestSemaphoreSelf(BaseTestSemaphore, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


@unittest.skipMPI("msmpi")
class TestSemaphoreWorld(BaseTestSemaphore, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


# ---

try:
    MPI.Win.Allocate(1, 1, MPI.INFO_NULL, MPI.COMM_SELF).Free()
except (NotImplementedError, MPI.Exception):
    unittest.skip("mpi-win-allocate")(BaseTestCounter)
    unittest.skip("mpi-win-allocate")(BaseTestMutexBasic)
    unittest.skip("mpi-win-allocate")(BaseTestMutexRecursive)
    unittest.skip("mpi-win-allocate")(BaseTestCondition)
    unittest.skip("mpi-win-allocate")(BaseTestSemaphore)

# ---

if __name__ == "__main__":
    unittest.main()



from mpi4py import MPI

try:
    import numpy
except ImportError:
    raise SystemExit from None

# transpose a matrix a into b

a = numpy.empty((100, 100), dtype=float, order="f")
b = numpy.empty((100, 100), dtype=float, order="f")
a.flat = numpy.arange(a.size, dtype=float)

lb, sizeofdouble = MPI.DOUBLE.Get_extent()

# create datatype dor one row
# (vector with 100 double entries and stride 100)
row = MPI.DOUBLE.Create_vector(100, 1, 100)

# create datatype for matrix in row-major order

# (one hundred copies of the row datatype, strided one word
#  apart; the successive row datatypes are interleaved)
xpose = row.Create_hvector(100, 1, sizeofdouble)
xpose.Commit()

# send matrix in row-major order and receive in column major order
abuf = (a, xpose)
bbuf = (b, MPI.DOUBLE)
myrank = MPI.COMM_WORLD.Get_rank()
status = MPI.Status()
MPI.COMM_WORLD.Sendrecv(abuf, myrank, 0, bbuf, myrank, 0, status)

assert numpy.allclose(a, b.transpose())
assert status.Get_count(xpose) == 1
assert status.Get_count(MPI.DOUBLE) == b.size

row.Free()
xpose.Free()



# -----------------------------------------------------------------------------

cdef object Lock  = None
cdef object RLock = None

try:
    from _thread import allocate_lock as Lock
    from _thread import RLock as RLock
except ImportError:                       # ~> uncovered
    from threading import Lock  as Lock   # ~> uncovered
    from threading import RLock as RLock  # ~> uncovered

# -----------------------------------------------------------------------------

cdef inline object New(type cls):
    return cls.__new__(cls)

# -----------------------------------------------------------------------------

ctypedef fused handle_t:
    MPI_Datatype
    MPI_Request
    MPI_Message
    MPI_Op
    MPI_Group
    MPI_Info
    MPI_Errhandler
    MPI_Session
    MPI_Comm
    MPI_Win
    MPI_File

cdef inline handle_t mpinull(handle_t _) noexcept nogil:
    cdef handle_t null
    if handle_t is MPI_Datatype   : null = MPI_DATATYPE_NULL
    if handle_t is MPI_Request    : null = MPI_REQUEST_NULL
    if handle_t is MPI_Message    : null = MPI_MESSAGE_NULL
    if handle_t is MPI_Op         : null = MPI_OP_NULL
    if handle_t is MPI_Group      : null = MPI_GROUP_NULL
    if handle_t is MPI_Info       : null = MPI_INFO_NULL
    if handle_t is MPI_Errhandler : null = MPI_ERRHANDLER_NULL
    if handle_t is MPI_Session    : null = MPI_SESSION_NULL
    if handle_t is MPI_Comm       : null = MPI_COMM_NULL
    if handle_t is MPI_Win        : null = MPI_WIN_NULL
    if handle_t is MPI_File       : null = MPI_FILE_NULL
    return null

cdef inline int named_Datatype(MPI_Datatype arg) noexcept nogil:
    if arg == MPI_DATATYPE_NULL           : return 1
    if arg == MPI_PACKED                  : return 1
    if arg == MPI_BYTE                    : return 1
    if arg == MPI_AINT                    : return 1
    if arg == MPI_OFFSET                  : return 1
    if arg == MPI_COUNT                   : return 1
    if arg == MPI_CHAR                    : return 1
    if arg == MPI_WCHAR                   : return 1
    if arg == MPI_SIGNED_CHAR             : return 1
    if arg == MPI_SHORT                   : return 1
    if arg == MPI_INT                     : return 1
    if arg == MPI_LONG                    : return 1
    if arg == MPI_LONG_LONG               : return 1
    if arg == MPI_UNSIGNED_CHAR           : return 1
    if arg == MPI_UNSIGNED_SHORT          : return 1
    if arg == MPI_UNSIGNED                : return 1
    if arg == MPI_UNSIGNED_LONG           : return 1
    if arg == MPI_UNSIGNED_LONG_LONG      : return 1
    if arg == MPI_FLOAT                   : return 1
    if arg == MPI_DOUBLE                  : return 1
    if arg == MPI_LONG_DOUBLE             : return 1
    if arg == MPI_C_BOOL                  : return 1
    if arg == MPI_INT8_T                  : return 1
    if arg == MPI_INT16_T                 : return 1
    if arg == MPI_INT32_T                 : return 1
    if arg == MPI_INT64_T                 : return 1
    if arg == MPI_UINT8_T                 : return 1
    if arg == MPI_UINT16_T                : return 1
    if arg == MPI_UINT32_T                : return 1
    if arg == MPI_UINT64_T                : return 1
    if arg == MPI_FLOAT16_T               : return 1
    if arg == MPI_BFLOAT16_T              : return 1
    if arg == MPI_C_COMPLEX               : return 1
    if arg == MPI_C_FLOAT_COMPLEX         : return 1
    if arg == MPI_C_DOUBLE_COMPLEX        : return 1
    if arg == MPI_C_LONG_DOUBLE_COMPLEX   : return 1
    if arg == MPI_CXX_BOOL                : return 1
    if arg == MPI_CXX_FLOAT_COMPLEX       : return 1
    if arg == MPI_CXX_DOUBLE_COMPLEX      : return 1
    if arg == MPI_CXX_LONG_DOUBLE_COMPLEX : return 1
    if arg == MPI_SHORT_INT               : return 1
    if arg == MPI_2INT                    : return 1
    if arg == MPI_LONG_INT                : return 1
    if arg == MPI_FLOAT_INT               : return 1
    if arg == MPI_DOUBLE_INT              : return 1
    if arg == MPI_LONG_DOUBLE_INT         : return 1
    if arg == MPI_CHARACTER               : return 1
    if arg == MPI_LOGICAL                 : return 1
    if arg == MPI_INTEGER                 : return 1
    if arg == MPI_REAL                    : return 1
    if arg == MPI_DOUBLE_PRECISION        : return 1
    if arg == MPI_COMPLEX                 : return 1
    if arg == MPI_DOUBLE_COMPLEX          : return 1
    if arg == MPI_LOGICAL1                : return 1
    if arg == MPI_LOGICAL2                : return 1
    if arg == MPI_LOGICAL4                : return 1
    if arg == MPI_LOGICAL8                : return 1
    if arg == MPI_LOGICAL16               : return 1
    if arg == MPI_INTEGER1                : return 1
    if arg == MPI_INTEGER2                : return 1
    if arg == MPI_INTEGER4                : return 1
    if arg == MPI_INTEGER8                : return 1
    if arg == MPI_INTEGER16               : return 1
    if arg == MPI_REAL2                   : return 1
    if arg == MPI_REAL4                   : return 1
    if arg == MPI_REAL8                   : return 1
    if arg == MPI_REAL16                  : return 1
    if arg == MPI_COMPLEX4                : return 1
    if arg == MPI_COMPLEX8                : return 1
    if arg == MPI_COMPLEX16               : return 1
    if arg == MPI_COMPLEX32               : return 1
    return 0

cdef inline int predef_Datatype(MPI_Datatype arg) noexcept nogil:
    if named_Datatype(arg): return 1
    cdef MPI_Count ni = 0, na = 0, nc = 0, nd = 0
    cdef int combiner = MPI_UNDEFINED
    cdef int ierr = MPI_Type_get_envelope_c(
        arg, &ni, &na, &nc, &nd, &combiner)
    if ierr != MPI_SUCCESS: return 0  # XXX Error?
    return (
        combiner == MPI_COMBINER_NAMED       or
        combiner == MPI_COMBINER_VALUE_INDEX or
        combiner == MPI_COMBINER_F90_INTEGER or
        combiner == MPI_COMBINER_F90_REAL    or
        combiner == MPI_COMBINER_F90_COMPLEX
    )

cdef inline int predef_Request(MPI_Request arg) noexcept nogil:
    if arg == MPI_REQUEST_NULL : return 1
    return 0

cdef inline int predef_Message(MPI_Message arg) noexcept nogil:
    if arg == MPI_MESSAGE_NULL    : return 1
    if arg == MPI_MESSAGE_NO_PROC : return 1
    return 0

cdef inline int predef_Op(MPI_Op arg) noexcept nogil:
    if arg == MPI_OP_NULL : return 1
    if arg == MPI_MAX     : return 1
    if arg == MPI_MIN     : return 1
    if arg == MPI_SUM     : return 1
    if arg == MPI_PROD    : return 1
    if arg == MPI_LAND    : return 1
    if arg == MPI_BAND    : return 1
    if arg == MPI_LOR     : return 1
    if arg == MPI_BOR     : return 1
    if arg == MPI_LXOR    : return 1
    if arg == MPI_BXOR    : return 1
    if arg == MPI_MAXLOC  : return 1
    if arg == MPI_MINLOC  : return 1
    if arg == MPI_REPLACE : return 1
    if arg == MPI_NO_OP   : return 1
    return 0

cdef inline int predef_Group(MPI_Group arg) noexcept nogil:
    if arg == MPI_GROUP_NULL  : return 1
    if arg == MPI_GROUP_EMPTY : return 1
    return 0

cdef inline int predef_Info(MPI_Info arg) noexcept nogil:
    if arg == MPI_INFO_NULL : return 1
    if arg == MPI_INFO_ENV  : return 1
    return 0

cdef inline int predef_Errhandler(MPI_Errhandler arg) noexcept nogil:
    if arg == MPI_ERRHANDLER_NULL  : return 1
    if arg == MPI_ERRORS_RETURN    : return 1
    if arg == MPI_ERRORS_ABORT     : return 1
    if arg == MPI_ERRORS_ARE_FATAL : return 1
    return 0

cdef inline int predef_Session(MPI_Session arg) noexcept nogil:
    if arg == MPI_SESSION_NULL : return 1
    return 0

cdef inline int predef_Comm(MPI_Comm arg) noexcept nogil:
    if arg == MPI_COMM_NULL  : return 1
    if arg == MPI_COMM_SELF  : return 1
    if arg == MPI_COMM_WORLD : return 1
    return 0

cdef inline int predef_Win(MPI_Win arg) noexcept nogil:
    if arg == MPI_WIN_NULL : return 1
    return 0

cdef inline int predef_File(MPI_File arg) noexcept nogil:
    if arg == MPI_FILE_NULL : return 1
    return 0

cdef inline int predefined(handle_t arg) noexcept nogil:
    cdef int result = 0
    if handle_t is MPI_Datatype   : result = predef_Datatype(arg)
    if handle_t is MPI_Request    : result = predef_Request(arg)
    if handle_t is MPI_Message    : result = predef_Message(arg)
    if handle_t is MPI_Op         : result = predef_Op(arg)
    if handle_t is MPI_Group      : result = predef_Group(arg)
    if handle_t is MPI_Info       : result = predef_Info(arg)
    if handle_t is MPI_Errhandler : result = predef_Errhandler(arg)
    if handle_t is MPI_Session    : result = predef_Session(arg)
    if handle_t is MPI_Comm       : result = predef_Comm(arg)
    if handle_t is MPI_Win        : result = predef_Win(arg)
    if handle_t is MPI_File       : result = predef_File(arg)
    return result

cdef inline int named(handle_t arg) noexcept nogil:
    if handle_t is MPI_Datatype:
        return named_Datatype(arg)
    else:
        return predefined(arg)

# -----------------------------------------------------------------------------

ctypedef fused PyMPIClass:
    Datatype
    Request
    Message
    Op
    Group
    Info
    Errhandler
    Session
    Comm
    Win
    File

cdef extern from * nogil:
    """
    #define PyMPI_FLAGS_READY   (1U<<0)
    #define PyMPI_FLAGS_CONST   (1U<<1)
    #define PyMPI_FLAGS_TEMP    (1U<<2)
    """
    enum: PyMPI_FLAGS_READY
    enum: PyMPI_FLAGS_CONST
    enum: PyMPI_FLAGS_TEMP

cdef inline int cinit(PyMPIClass self, PyMPIClass arg) except -1:
    self.ob_mpi = mpinull(self.ob_mpi)
    self.flags |= PyMPI_FLAGS_READY
    if arg is None: return 0
    self.ob_mpi = arg.ob_mpi
    if PyMPIClass is Request:
        self.ob_buf = arg.ob_buf
    if PyMPIClass is Message:
        self.ob_buf = arg.ob_buf
    if PyMPIClass is Op:
        op_user_cpy(self, arg)
    if PyMPIClass is Win:
        self.ob_mem = arg.ob_mem
    return 0

cdef inline int marktemp(PyMPIClass self) except -1:
    if not predefined(self.ob_mpi):
        self.flags |= PyMPI_FLAGS_TEMP
    return 0


@cython.linetrace(False)
cdef inline int freetemp(PyMPIClass self) except -1:
    if named(self.ob_mpi): return 0
    if not mpi_active(): return 0
    if PyMPIClass is Datatype:
        CHKERR( MPI_Type_free(&self.ob_mpi) )
    return 0


@cython.linetrace(False)
cdef inline int dealloc(PyMPIClass self) except -1:
    if not (self.flags & PyMPI_FLAGS_READY): return 0
    if (self.flags & PyMPI_FLAGS_CONST): return 0
    if (self.flags & PyMPI_FLAGS_TEMP ): return freetemp(self)
    if self.flags: return 0  # TODO: this always return

    if not mpi_active(): return 0
    if predefined(self.ob_mpi): return 0
    PyErr_WarnFormat(
        RuntimeWarning, 1,
        b"collecting object with %.200U handle %p",
        <PyObject*> cython.typeof(self.ob_mpi),
        <void*> <Py_uintptr_t> self.ob_mpi,
    )

cdef extern from "Python.h":
    enum: Py_LT
    enum: Py_LE
    enum: Py_EQ
    enum: Py_NE
    enum: Py_GT
    enum: Py_GE

cdef inline object richcmp(PyMPIClass self, object other, int op):
    if op == Py_EQ: return (self.ob_mpi == (<PyMPIClass>other).ob_mpi)
    if op == Py_NE: return (self.ob_mpi != (<PyMPIClass>other).ob_mpi)
    cdef str mod = type(self).__module__
    cdef str cls = type(self).__name__
    raise TypeError(f"unorderable type '{mod}.{cls}'")

cdef inline int nonnull(PyMPIClass self) noexcept nogil:
    return self.ob_mpi != mpinull(self.ob_mpi)

cdef inline int constobj(PyMPIClass self) noexcept nogil:
    return self.flags & PyMPI_FLAGS_CONST

# -----------------------------------------------------------------------------

cdef dict def_registry = {}

cdef inline type def_class(handle_t handle):
    <void> handle  # unused
    cdef type result = None
    if handle_t is MPI_Datatype   : result = Datatype
    if handle_t is MPI_Request    : result = Request
    if handle_t is MPI_Message    : result = Message
    if handle_t is MPI_Op         : result = Op
    if handle_t is MPI_Group      : result = Group
    if handle_t is MPI_Info       : result = Info
    if handle_t is MPI_Errhandler : result = Errhandler
    if handle_t is MPI_Session    : result = Session
    if handle_t is MPI_Comm       : result = Comm
    if handle_t is MPI_Win        : result = Win
    if handle_t is MPI_File       : result = File
    return result

cdef inline int def_register(
    handle_t handle,
    object   pyobj,
    object   name,
) except -1:
    cdef type cls = def_class(handle)
    cdef dict registry = def_registry.get(cls)
    cdef object key = <Py_uintptr_t> handle
    if registry is None:
        registry = def_registry[cls] = {}
    if key not in registry:
        registry[key] = (pyobj, name)
    return 0

cdef inline object def_lookup(handle_t handle):
    cdef type cls = def_class(handle)
    cdef dict registry = def_registry[cls]
    cdef object key = <Py_uintptr_t> handle
    return registry[key]

cdef __newobj__ = None
from copyreg import __newobj__

cdef inline object def_reduce(PyMPIClass self):
    cdef object pyobj, name
    pyobj, name = def_lookup(self.ob_mpi)
    if self is pyobj: return name
    return (__newobj__, (type(self), pyobj))

cdef inline object reduce_default(PyMPIClass self):
    if named(self.ob_mpi):
        return def_reduce(self)
    cdef str mod = type(self).__module__
    cdef str cls = type(self).__name__
    raise ValueError(f"cannot serialize '{mod}.{cls}' instance")

# -----------------------------------------------------------------------------

cdef inline Py_uintptr_t tohandle(PyMPIClass self) noexcept nogil:
    return <Py_uintptr_t> self.ob_mpi

cdef inline object fromhandle(handle_t arg):
    cdef object obj = None
    if handle_t is MPI_Datatype   : obj = PyMPIDatatype_New(arg)
    if handle_t is MPI_Request    : obj = PyMPIRequest_New(arg)
    if handle_t is MPI_Message    : obj = PyMPIMessage_New(arg)
    if handle_t is MPI_Op         : obj = PyMPIOp_New(arg)
    if handle_t is MPI_Group      : obj = PyMPIGroup_New(arg)
    if handle_t is MPI_Info       : obj = PyMPIInfo_New(arg)
    if handle_t is MPI_Errhandler : obj = PyMPIErrhandler_New(arg)
    if handle_t is MPI_Session    : obj = PyMPISession_New(arg)
    if handle_t is MPI_Comm       : obj = PyMPIComm_New(arg)
    if handle_t is MPI_Win        : obj = PyMPIWin_New(arg)
    if handle_t is MPI_File       : obj = PyMPIFile_New(arg)
    return obj

# -----------------------------------------------------------------------------

cdef inline int nullify(PyMPIClass self) except -1:
    self.ob_mpi = mpinull(self.ob_mpi)
    return 0

cdef inline int callfree(PyMPIClass self) except -1:
    if PyMPIClass is not Errhandler:
        if predefined(self.ob_mpi):
            return nullify(self)
    if PyMPIClass is Datatype:
        self.Free()
    if PyMPIClass is Request:
        self.Free()
    if PyMPIClass is Message:
        pass
    if PyMPIClass is Op:
        self.Free()
    if PyMPIClass is Group:
        self.Free()
    if PyMPIClass is Info:
        self.Free()
    if PyMPIClass is Errhandler:
        self.Free()
    if PyMPIClass is Session:
        self.Finalize()
    if PyMPIClass is Comm:
        self.Free()
    if PyMPIClass is Win:
        self.Free()
    if PyMPIClass is File:
        self.Close()
    return 0

cdef inline int safefree(PyMPIClass self) except -1:
    # skip freeing module constant objects
    if self.flags & PyMPI_FLAGS_CONST:
        return 0
    # skip freeing objects with null handles
    if self.ob_mpi == mpinull(self.ob_mpi):
        return 0
    # since MPI-4, some objects can be freeded
    # before/after the world model init/finalize
    if (
        PyMPIClass is Info or
        PyMPIClass is Session or
        PyMPIClass is Errhandler
    ) and mpi_version >= 4:
        return callfree(self)
    # skip freeing before/after init/finalize
    if not mpi_active():
        return nullify(self)  # ~> uncovered
    # can safely free object
    return callfree(self)

# -----------------------------------------------------------------------------

# Status

cdef inline MPI_Status *arg_Status(object status) except? NULL:
    if status is None: return MPI_STATUS_IGNORE
    return &((<Status?>status).ob_mpi)

# -----------------------------------------------------------------------------

# Datatype

cdef inline Datatype def_Datatype(MPI_Datatype arg, object name):
    cdef Datatype obj = Datatype.__new__(Datatype)
    obj.ob_mpi = arg
    obj.flags |= PyMPI_FLAGS_CONST
    def_register(arg, obj, name)
    return obj

cdef inline Datatype ref_Datatype(MPI_Datatype arg):
    cdef Datatype obj = Datatype.__new__(Datatype)
    obj.ob_mpi = arg
    if not predefined(arg):
        obj.flags |= 0  # TODO
    return obj

cdef inline object reduce_Datatype(Datatype self):
    # named
    if named(self.ob_mpi):
        return def_reduce(self)
    # predefined and user-defined
    cdef object basetype, combiner, params
    basetype, combiner, params = datatype_decode(self, True)
    return (_datatype_create, (basetype, combiner, params, True))

# -----------------------------------------------------------------------------

# Request

cdef inline Request def_Request(MPI_Request arg, object name):
    cdef Request obj = Request.__new__(Request)
    obj.ob_mpi = arg
    obj.flags |= PyMPI_FLAGS_CONST
    def_register(arg, obj, name)
    return obj

# -----------------------------------------------------------------------------

# Message

cdef inline Message def_Message(MPI_Message arg, object name):
    cdef Message obj = Message.__new__(Message)
    obj.ob_mpi = arg
    obj.flags |= PyMPI_FLAGS_CONST
    def_register(arg, obj, name)
    return obj

# -----------------------------------------------------------------------------

# Op

cdef dict def_op = {}

cdef inline Op def_Op(MPI_Op arg, object name):
    cdef Op obj = Op.__new__(Op)
    obj.ob_mpi = arg
    obj.flags |= PyMPI_FLAGS_CONST
    def_register(arg, obj, name)
    return obj

cdef inline object reduce_Op(Op self):
    # predefined
    if named(self.ob_mpi):
        return def_reduce(self)
    # user-defined
    cdef int index = op_user_id_get(self)
    if index == 0:
        raise ValueError("cannot pickle user-defined reduction operation")
    cdef object function = op_user_registry[index]
    cdef object commute = self.Is_commutative()
    return (type(self).Create, (function, commute,))

# -----------------------------------------------------------------------------

# Group

cdef inline Group def_Group(MPI_Group arg, object name):
    cdef Group obj = Group.__new__(Group)
    obj.ob_mpi = arg
    obj.flags |= PyMPI_FLAGS_CONST
    def_register(arg, obj, name)
    return obj

# -----------------------------------------------------------------------------

# Info

cdef inline Info def_Info(MPI_Info arg, object name):
    cdef Info obj = Info.__new__(Info)
    obj.ob_mpi = arg
    obj.flags |= PyMPI_FLAGS_CONST
    def_register(arg, obj, name)
    return obj


cdef extern from *:
    const MPI_Info _info_null "MPI_INFO_NULL"

cdef inline MPI_Info arg_Info(object obj) except? _info_null:
    if obj is None: return MPI_INFO_NULL
    return (<Info?>obj).ob_mpi

cdef inline object reduce_Info(Info self):
    # predefined
    if named(self.ob_mpi):
        return def_reduce(self)
    # user-defined
    return (type(self).Create, (self.items(),))

# -----------------------------------------------------------------------------

# Errhandler

cdef inline Errhandler def_Errhandler(MPI_Errhandler arg, object name):
    cdef Errhandler obj = Errhandler.__new__(Errhandler)
    obj.ob_mpi = arg
    obj.flags |= PyMPI_FLAGS_CONST
    def_register(arg, obj, name)
    return obj

cdef extern from *:
    const MPI_Errhandler _errhandler_null "MPI_ERRHANDLER_NULL"

cdef inline MPI_Errhandler arg_Errhandler(object obj) except? _errhandler_null:
    if obj is not None: return (<Errhandler?>obj).ob_mpi
    cdef MPI_Errhandler errhandler = options_get_errhandler()
    if errhandler == MPI_ERRHANDLER_NULL: errhandler = MPI_ERRORS_ARE_FATAL
    return errhandler

# -----------------------------------------------------------------------------

# Session

cdef inline Session def_Session(MPI_Session arg, object name):
    cdef Session obj = Session.__new__(Session)
    obj.ob_mpi = arg
    obj.flags |= PyMPI_FLAGS_CONST
    def_register(arg, obj, name)
    return obj

# -----------------------------------------------------------------------------

# Comm

cdef inline type CommType(MPI_Comm arg):
    if arg == MPI_COMM_NULL:
        return Comm
    if arg == MPI_COMM_SELF:
        return Intracomm
    if arg == MPI_COMM_WORLD:
        return Intracomm
    cdef int inter = 0
    CHKERR( MPI_Comm_test_inter(arg, &inter) )
    if inter:
        return Intercomm
    cdef int topo  = MPI_UNDEFINED
    CHKERR( MPI_Topo_test(arg, &topo) )
    if topo == MPI_UNDEFINED:
        return Intracomm
    if topo == MPI_CART:
        return Cartcomm
    if topo == MPI_GRAPH:
        return Graphcomm
    if topo == MPI_DIST_GRAPH:
        return Distgraphcomm
    return Comm  # ~> unreachable

cdef inline Comm def_Comm(MPI_Comm arg, object name):
    cdef Comm obj = Comm.__new__(Comm)
    obj.ob_mpi = arg
    obj.flags |= PyMPI_FLAGS_CONST
    def_register(arg, obj, name)
    return obj

cdef inline Intracomm def_Intracomm(MPI_Comm arg, object name):
    cdef Intracomm obj = Intracomm.__new__(Intracomm)
    obj.ob_mpi = arg
    obj.flags |= PyMPI_FLAGS_CONST
    def_register(arg, obj, name)
    return obj

cdef inline Intercomm def_Intercomm(MPI_Comm arg):
    cdef Intercomm obj = Intercomm.__new__(Intercomm)
    obj.ob_mpi = arg
    obj.flags |= PyMPI_FLAGS_CONST
    return obj

# -----------------------------------------------------------------------------

# Win

cdef inline Win def_Win(MPI_Win arg, object name):
    cdef Win obj = Win.__new__(Win)
    obj.ob_mpi = arg
    obj.flags |= PyMPI_FLAGS_CONST
    def_register(arg, obj, name)
    return obj

# -----------------------------------------------------------------------------

# File

cdef inline File def_File(MPI_File arg, object name):
    cdef File obj = File.__new__(File)
    obj.ob_mpi = arg
    obj.flags |= PyMPI_FLAGS_CONST
    def_register(arg, obj, name)
    return obj

# -----------------------------------------------------------------------------



import unittest

from reductions import Intracomm

from mpi4py import MPI


class BaseTest:
    def test_reduce(self):
        rank = self.comm.rank
        size = self.comm.size
        for root in range(size):
            msg = rank
            res = self.comm.reduce(sendobj=msg, root=root)
            if self.comm.rank == root:
                self.assertEqual(res, sum(range(size)))
            else:
                self.assertIsNone(res)

    def test_reduce_min(self):
        rank = self.comm.rank
        size = self.comm.size
        for root in range(size):
            msg = rank
            res = self.comm.reduce(sendobj=msg, op=MPI.MIN, root=root)
            if self.comm.rank == root:
                self.assertEqual(res, 0)
            else:
                self.assertIsNone(res)

    def test_reduce_max(self):
        rank = self.comm.rank
        size = self.comm.size
        for root in range(size):
            msg = rank
            res = self.comm.reduce(sendobj=msg, op=MPI.MAX, root=root)
            if self.comm.rank == root:
                self.assertEqual(res, size - 1)
            else:
                self.assertIsNone(res)

    def test_reduce_minloc(self):
        rank = self.comm.rank
        size = self.comm.size
        for root in range(size):
            msg = rank
            res = self.comm.reduce(
                sendobj=(msg, rank), op=MPI.MINLOC, root=root
            )
            if self.comm.rank == root:
                self.assertEqual(res, (0, 0))
            else:
                self.assertIsNone(res)

    def test_reduce_maxloc(self):
        rank = self.comm.rank
        size = self.comm.size
        for root in range(size):
            msg = rank
            res = self.comm.reduce(
                sendobj=(msg, rank), op=MPI.MAXLOC, root=root
            )
            if self.comm.rank == root:
                self.assertEqual(res, (size - 1, size - 1))
            else:
                self.assertIsNone(res)

    def test_allreduce(self):
        rank = self.comm.rank
        size = self.comm.size
        msg = rank
        res = self.comm.allreduce(sendobj=msg)
        self.assertEqual(res, sum(range(size)))

    def test_allreduce_min(self):
        rank = self.comm.rank
        msg = rank
        res = self.comm.allreduce(sendobj=msg, op=MPI.MIN)
        self.assertEqual(res, 0)

    def test_allreduce_max(self):
        rank = self.comm.rank
        size = self.comm.size
        msg = rank
        res = self.comm.allreduce(sendobj=msg, op=MPI.MAX)
        self.assertEqual(res, size - 1)

    def test_allreduce_minloc(self):
        rank = self.comm.rank
        msg = rank
        res = self.comm.allreduce(sendobj=(msg, rank), op=MPI.MINLOC)
        self.assertEqual(res, (0, 0))

    def test_allreduce_maxloc(self):
        rank = self.comm.rank
        size = self.comm.size
        msg = rank
        res = self.comm.allreduce(sendobj=(msg, rank), op=MPI.MAXLOC)
        self.assertEqual(res, (size - 1, size - 1))

    def test_scan(self):
        rank = self.comm.rank
        size = self.comm.size
        msg = rank
        res = self.comm.scan(sendobj=msg)
        self.assertEqual(res, sum(list(range(size))[: rank + 1]))

    def test_scan_min(self):
        rank = self.comm.rank
        msg = rank
        res = self.comm.scan(sendobj=msg, op=MPI.MIN)
        self.assertEqual(res, 0)

    def test_scan_max(self):
        rank = self.comm.rank
        msg = rank
        res = self.comm.scan(sendobj=msg, op=MPI.MAX)
        self.assertEqual(res, rank)

    def test_scan_minloc(self):
        rank = self.comm.rank
        msg = rank
        res = self.comm.scan(sendobj=(msg, rank), op=MPI.MINLOC)
        self.assertEqual(res, (0, 0))

    def test_scan_maxloc(self):
        rank = self.comm.rank
        msg = rank
        res = self.comm.scan(sendobj=(msg, rank), op=MPI.MAXLOC)
        self.assertEqual(res, (rank, rank))

    def test_exscan(self):
        rank = self.comm.rank
        size = self.comm.size
        msg = rank
        res = self.comm.exscan(sendobj=msg)
        if self.comm.rank == 0:
            self.assertIsNone(res)
        else:
            self.assertEqual(res, sum(list(range(size))[:rank]))

    def test_exscan_min(self):
        rank = self.comm.rank
        msg = rank
        res = self.comm.exscan(sendobj=msg, op=MPI.MIN)
        if self.comm.rank == 0:
            self.assertIsNone(res)
        else:
            self.assertEqual(res, 0)

    def test_exscan_max(self):
        rank = self.comm.rank
        msg = rank
        res = self.comm.exscan(sendobj=msg, op=MPI.MAX)
        if self.comm.rank == 0:
            self.assertIsNone(res)
        else:
            self.assertEqual(res, rank - 1)

    def test_exscan_minloc(self):
        rank = self.comm.rank
        msg = rank
        res = self.comm.exscan(sendobj=(msg, rank), op=MPI.MINLOC)
        if self.comm.rank == 0:
            self.assertIsNone(res)
        else:
            self.assertEqual(res, (0, 0))

    def test_exscan_maxloc(self):
        rank = self.comm.rank
        msg = rank
        res = self.comm.exscan(sendobj=(msg, rank), op=MPI.MAXLOC)
        if self.comm.rank == 0:
            self.assertIsNone(res)
        else:
            self.assertEqual(res, (rank - 1, rank - 1))


class TestS(BaseTest, unittest.TestCase):
    def setUp(self):
        self.comm = Intracomm(MPI.COMM_SELF)


class TestW(BaseTest, unittest.TestCase):
    def setUp(self):
        self.comm = Intracomm(MPI.COMM_WORLD)


class TestSD(BaseTest, unittest.TestCase):
    def setUp(self):
        self.comm = Intracomm(MPI.COMM_SELF.Dup())

    def tearDown(self):
        self.comm.Free()


class TestWD(BaseTest, unittest.TestCase):
    def setUp(self):
        self.comm = Intracomm(MPI.COMM_WORLD.Dup())

    def tearDown(self):
        self.comm.Free()


if __name__ == "__main__":
    unittest.main()



# -----------------------------------------------------------------------------

# From dlpack.h (as of v0.8)

cdef extern from * nogil:
    ctypedef unsigned char      uint8_t
    ctypedef unsigned short     uint16_t
    ctypedef signed   int       int32_t
    ctypedef unsigned int       uint32_t
    ctypedef signed   long long int64_t
    ctypedef unsigned long long uint64_t

ctypedef struct DLPackVersion:
    uint32_t major
    uint32_t minor

ctypedef enum DLDeviceType:
    kDLCPU = 1
    kDLCUDA = 2
    kDLCUDAHost = 3
    kDLOpenCL = 4
    kDLVulkan = 7
    kDLMetal = 8
    kDLVPI = 9
    kDLROCM = 10
    kDLROCMHost = 11
    kDLExtDev = 12
    kDLCUDAManaged = 13
    kDLOneAPI = 14
    kDLWebGPU = 15
    kDLHexagon = 16
    kDLMAIA = 17

ctypedef struct DLDevice:
    DLDeviceType device_type
    int32_t device_id

ctypedef enum DLDataTypeCode:
    kDLInt = 0
    kDLUInt = 1
    kDLFloat = 2
    kDLOpaqueHandle = 3
    kDLBfloat = 4
    kDLComplex = 5
    kDLBool = 6

ctypedef struct DLDataType:
    uint8_t code
    uint8_t bits
    uint16_t lanes

ctypedef struct DLTensor:
    void *data
    DLDevice device
    int32_t ndim
    DLDataType dtype
    int64_t *shape
    int64_t *strides
    uint64_t byte_offset

ctypedef struct DLManagedTensor:
    DLTensor dl_tensor
    void *manager_ctx
    void (*deleter)(DLManagedTensor *)

ctypedef enum:
    DLPACK_FLAG_BITMASK_READ_ONLY = (1UL << 0UL)
    DLPACK_FLAG_BITMASK_IS_COPIED = (1UL << 1UL)

ctypedef struct DLManagedTensorVersioned:
    DLPackVersion version
    void *manager_ctx
    void (*deleter)(DLManagedTensorVersioned *)
    uint64_t flags
    DLTensor dl_tensor

# -----------------------------------------------------------------------------

cdef extern from "Python.h":
    void* PyCapsule_GetPointer(object, const char[]) except? NULL
    int PyCapsule_SetName(object, const char[]) except -1
    int PyCapsule_IsValid(object, const char[])

# -----------------------------------------------------------------------------

cdef inline int dlpack_check_version(
    const DLPackVersion *version,
    unsigned version_major,
) except -1:
    if version == NULL: return 0
    if version.major >= version_major: return 0
    raise BufferError("dlpack: unexpected version")

cdef inline int dlpack_is_contig(
    const DLTensor *dltensor,
    char order,
) noexcept nogil:
    cdef int i, ndim = dltensor.ndim
    cdef int64_t *shape = dltensor.shape
    cdef int64_t *strides = dltensor.strides
    cdef int64_t start, step, index, dim, size = 1
    if strides == NULL:
        return order != c'F' or ndim <= 1
    if order == c'F':
        start = 0
        step = 1
    else:
        start = ndim - 1
        step = -1
    for i in range(ndim):
        index = start + step * i
        dim = shape[index]
        if dim > 1 and size != strides[index]:
            return 0
        size *= dim
    return 1

cdef inline int dlpack_check_shape(const DLTensor *dltensor) except -1:
    cdef int i, ndim = dltensor.ndim
    if ndim < 0:
        raise BufferError("dlpack: number of dimensions is negative")
    if ndim > 0 and dltensor.shape == NULL:
        raise BufferError("dlpack: shape is NULL")
    for i in range(ndim):
        if dltensor.shape[i] < 0:
            raise BufferError("dlpack: shape item is negative")
    if dltensor.strides != NULL:
        for i in range(ndim):
            if dltensor.strides[i] < 0:
                raise BufferError("dlpack: strides item is negative")
    return 0

cdef inline int dlpack_check_contig(const DLTensor *dltensor) except -1:
    cdef bint c_contig = dlpack_is_contig(dltensor, c'C')
    cdef bint f_contig = dlpack_is_contig(dltensor, c'F')
    if not c_contig and not f_contig:
        raise BufferError("dlpack: buffer is not contiguous")
    return 0

cdef inline int dlpack_check_rdonly(bint readonly, int flags) except -1:
    if readonly and ((flags & PyBUF_WRITABLE) == PyBUF_WRITABLE):
        raise BufferError("dlpack: buffer is not writable")
    return 0

cdef inline void *dlpack_get_data(
    const DLTensor *dltensor,
) noexcept nogil:
    return <char*> dltensor.data + dltensor.byte_offset

cdef inline Py_ssize_t dlpack_get_size(
    const DLTensor *dltensor,
) noexcept nogil:
    cdef int i, ndim = dltensor.ndim
    cdef int64_t *shape = dltensor.shape
    cdef Py_ssize_t bits = dltensor.dtype.bits
    cdef Py_ssize_t lanes = dltensor.dtype.lanes
    cdef Py_ssize_t size = 1
    for i in range(ndim):
        size *= <Py_ssize_t> shape[i]
    size *= (bits * lanes + 7) // 8
    return size

cdef inline char *dlpack_get_format(
    const DLTensor *dltensor,
) noexcept nogil:
    cdef unsigned int code = dltensor.dtype.code
    cdef unsigned int bits = dltensor.dtype.bits
    if dltensor.dtype.lanes != 1:
        if code == kDLFloat and dltensor.dtype.lanes == 2:
            if bits == 8*sizeof(float):       return b"Zf"
            if bits == 8*sizeof(double):      return b"Zd"
            if bits == 8*sizeof(long double): return b"Zg"
        return b"B"
    if code == kDLBool:
        if bits == 8: return b"?"
    if code == kDLInt:
        if bits == 8*sizeof(char):      return b"b"
        if bits == 8*sizeof(short):     return b"h"
        if bits == 8*sizeof(int):       return b"i"
        if bits == 8*sizeof(long):      return b"l"
        if bits == 8*sizeof(long long): return b"q"  # ~> long
    if code == kDLUInt:
        if bits == 8*sizeof(char):      return b"B"
        if bits == 8*sizeof(short):     return b"H"
        if bits == 8*sizeof(int):       return b"I"
        if bits == 8*sizeof(long):      return b"L"
        if bits == 8*sizeof(long long): return b"Q"  # ~> long
    if code == kDLFloat:
        if bits ==  8*sizeof(float)//2:    return b"e"
        if bits ==  8*sizeof(float):       return b"f"
        if bits ==  8*sizeof(double):      return b"d"
        if bits ==  8*sizeof(long double): return b"g"
    if code == kDLComplex:
        if bits ==  8*2*sizeof(float)//2:    return b"Ze"
        if bits ==  8*2*sizeof(float):       return b"Zf"
        if bits ==  8*2*sizeof(double):      return b"Zd"
        if bits ==  8*2*sizeof(long double): return b"Zg"  # ~> uncovered
    return BYTE_FMT

cdef inline Py_ssize_t dlpack_get_itemsize(
    const DLTensor *dltensor,
) noexcept nogil:
    cdef unsigned int code = dltensor.dtype.code
    cdef unsigned int bits = dltensor.dtype.bits
    if dltensor.dtype.lanes != 1:
        if code == kDLFloat and dltensor.dtype.lanes == 2:
            if (
                bits == 8*sizeof(float) or
                bits == 8*sizeof(double) or
                bits == 8*sizeof(long double)
            ):
                return <Py_ssize_t> bits // 8 * 2
        bits = 1
    return <Py_ssize_t> (bits + 7) // 8

# -----------------------------------------------------------------------------

cdef int Py_CheckDLPackBuffer(object obj) noexcept:
    try:    return <bint>hasattr(obj, '__dlpack__')
    except: return 0  # ~> uncovered  # noqa

cdef int Py_GetDLPackBuffer(object obj, Py_buffer *view, int flags) except -1:
    cdef unsigned version_major = 1
    cdef const char *capsulename = b"dltensor_versioned"
    cdef uint64_t READONLY = DLPACK_FLAG_BITMASK_READ_ONLY
    cdef object dlpack
    cdef object dlpack_device
    cdef tuple max_version
    cdef unsigned device_type
    cdef int device_id
    cdef object capsule
    cdef void *pointer
    cdef DLManagedTensorVersioned *managed1 = NULL
    cdef DLManagedTensor *managed0 = NULL
    cdef const DLPackVersion *dlversion
    cdef const DLTensor *dltensor
    cdef void *buf
    cdef Py_ssize_t size
    cdef Py_ssize_t itemsize
    cdef char *format
    cdef bint readonly

    try:
        dlpack = obj.__dlpack__
        dlpack_device = obj.__dlpack_device__
    except AttributeError:
        raise NotImplementedError("dlpack: missing support")

    device_type, device_id = dlpack_device()
    <void> device_id  # unused

    try:  # DLPack v1.0+
        max_version = (version_major, 0)
        if device_type == kDLCPU:
            capsule = dlpack(max_version=max_version, copy=False)
        else:
            capsule = dlpack(stream=-1, max_version=max_version, copy=False)
    except TypeError:  # DLPack v0.x
        version_major = 0
        capsulename = b"dltensor"
        if device_type == kDLCPU:
            capsule = dlpack()
        else:
            capsule = dlpack(stream=-1)

    if not PyCapsule_IsValid(capsule, capsulename):
        raise BufferError("dlpack: invalid capsule object")
    pointer = PyCapsule_GetPointer(capsule, capsulename)
    if version_major >= 1:
        managed1  = <DLManagedTensorVersioned*> pointer
        dlversion = &managed1.version
        dltensor  = &managed1.dl_tensor
        readonly  = (managed1.flags & READONLY) == READONLY
    else:
        managed0  = <DLManagedTensor*> pointer
        dlversion = NULL
        dltensor  = &managed0.dl_tensor
        readonly  = 0

    dlpack_check_version(dlversion, version_major)
    dlpack_check_shape(dltensor)
    dlpack_check_contig(dltensor)
    dlpack_check_rdonly(readonly, flags)

    buf = dlpack_get_data(dltensor)
    size = dlpack_get_size(dltensor)
    itemsize = dlpack_get_itemsize(dltensor)
    format = dlpack_get_format(dltensor)

    PyBuffer_FillInfo(view, capsule, buf, size, readonly, flags)

    if (flags & PyBUF_FORMAT) == PyBUF_FORMAT:
        view.format = format
        if view.format != BYTE_FMT:
            view.itemsize = itemsize

    return <int> device_type

# -----------------------------------------------------------------------------



# Demonstrate how to work with Python GPU arrays using CUDA-aware MPI.
# A GPU array is allocated and manipulated through Numba, which is
# compliant with the __cuda_array_interface__ standard.
#
# Run this script using the following command:
# mpiexec -n 2 python use_cupy.py

import numpy
from numba import cuda

from mpi4py import MPI


@cuda.jit()
def add_const(arr, value):
    x = cuda.grid(1)
    if x < arr.size:
        arr[x] += value


comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()

# Send-Recv
if rank == 0:
    buf = cuda.device_array((20,), dtype="f")
    buf[:] = range(20)
    block = 32
    grid = (buf.size + block - 1) // block
    add_const[grid, block](buf, 100)
    # always make sure the GPU buffer is ready before any MPI operation
    cuda.default_stream().synchronize()
    comm.Send(buf, dest=1, tag=77)
else:
    buf = cuda.device_array((20,), dtype="f")
    cuda.default_stream().synchronize()
    comm.Recv(buf, source=0, tag=77)
    buf = buf.copy_to_host()
    assert numpy.allclose(buf, 100 + numpy.arange(20, dtype="f"))



import os
import pathlib
import tempfile

import mpiunittest as unittest

from mpi4py import MPI


def maketemp(prefix):
    fd, name = tempfile.mkstemp(prefix=prefix)
    os.close(fd)
    return name


class BaseTestFile:
    #
    COMM = MPI.COMM_NULL
    FILE = MPI.FILE_NULL

    prefix = "mpi4py"

    def setUp(self):
        self.fname = pathlib.Path(maketemp(self.prefix))
        self.amode = MPI.MODE_RDWR | MPI.MODE_CREATE
        # self.amode |= MPI.MODE_DELETE_ON_CLOSE
        try:
            self.FILE = MPI.File.Open(
                self.COMM,
                self.fname,
                self.amode,
                MPI.INFO_NULL,
            )
            # self.fname=None
        except Exception:
            self.fname.unlink()
            raise

    def tearDown(self):
        if self.FILE == MPI.FILE_NULL:
            return
        amode = self.FILE.amode
        self.FILE.Close()
        if not (amode & MPI.MODE_DELETE_ON_CLOSE):
            MPI.File.Delete(self.fname, MPI.INFO_NULL)

    @unittest.skipMPI("openmpi(==2.0.0)")
    @unittest.skipMPI("MPICH2(<1.1.0)")
    def testPreallocate(self):
        size = self.FILE.Get_size()
        self.assertEqual(size, 0)
        self.FILE.Preallocate(1)
        size = self.FILE.Get_size()
        self.assertEqual(size, 1)
        self.FILE.Preallocate(100)
        size = self.FILE.Get_size()
        self.assertEqual(size, 100)
        self.FILE.Preallocate(10)
        size = self.FILE.Get_size()
        self.assertEqual(size, 100)
        self.FILE.Preallocate(200)
        size = self.FILE.Get_size()
        self.assertEqual(size, 200)

    def testGetSetSize(self):
        size = self.FILE.Get_size()
        self.assertEqual(size, 0)
        size = self.FILE.size
        self.assertEqual(size, 0)
        self.FILE.Set_size(100)
        size = self.FILE.Get_size()
        self.assertEqual(size, 100)
        size = self.FILE.size
        self.assertEqual(size, 100)

    def testGetGroup(self):
        fgroup = self.FILE.Get_group()
        cgroup = self.COMM.Get_group()
        gcomp = MPI.Group.Compare(fgroup, cgroup)
        self.assertEqual(gcomp, MPI.IDENT)
        fgroup.Free()
        cgroup.Free()

    def testGetAmode(self):
        amode = self.FILE.Get_amode()
        self.assertEqual(self.amode, amode)
        self.assertEqual(self.FILE.amode, self.amode)

    def testGetSetInfo(self):
        # info = MPI.INFO_NULL
        # self.FILE.Set_info(info)
        info = MPI.Info.Create()
        self.FILE.Set_info(info)
        info.Free()
        info = self.FILE.Get_info()
        self.FILE.Set_info(info)
        info.Free()

    def testGetSetView(self):
        fsize = 100 * MPI.DOUBLE.size
        self.FILE.Set_size(fsize)
        displacements = range(100)
        datatypes = [MPI.SHORT, MPI.INT, MPI.LONG, MPI.FLOAT, MPI.DOUBLE]
        datareps = ["native"]  # ['native', 'internal', 'external32']
        for disp in displacements:
            for dtype in datatypes:
                for datarep in datareps:
                    etype, ftype = dtype, dtype
                    self.FILE.Set_view(
                        disp, etype, ftype, datarep, MPI.INFO_NULL
                    )
                    of, et, ft, dr = self.FILE.Get_view()
                    self.assertEqual(disp, of)
                    self.assertEqual(etype.Get_extent(), et.Get_extent())
                    self.assertEqual(ftype.Get_extent(), ft.Get_extent())
                    self.assertEqual(datarep, dr)
                    try:
                        if not et.is_predefined:
                            et.Free()
                    except NotImplementedError:
                        if et != etype:
                            et.Free()
                    try:
                        if not ft.is_predefined:
                            ft.Free()
                    except NotImplementedError:
                        if ft != ftype:
                            ft.Free()

    def testGetSetAtomicity(self):
        atom = self.FILE.Get_atomicity()
        self.assertFalse(atom)
        for atomicity in [True, False] * 4:
            self.FILE.Set_atomicity(atomicity)
            atom = self.FILE.Get_atomicity()
            self.assertEqual(atom, atomicity)

    def testSync(self):
        self.FILE.Sync()

    def testSeekGetPosition(self):
        offset = 0
        self.FILE.Seek(offset, MPI.SEEK_END)
        self.FILE.Seek(offset, MPI.SEEK_CUR)
        self.FILE.Seek(offset, MPI.SEEK_SET)
        pos = self.FILE.Get_position()
        self.assertEqual(pos, offset)

    def testSeekGetPositionShared(self):
        offset = 0
        self.FILE.Seek_shared(offset, MPI.SEEK_END)
        self.FILE.Seek_shared(offset, MPI.SEEK_CUR)
        self.FILE.Seek_shared(offset, MPI.SEEK_SET)
        pos = self.FILE.Get_position_shared()
        self.assertEqual(pos, offset)

    @unittest.skipMPI("openmpi(==2.0.0)")
    def testGetByteOffset(self):
        for offset in range(10):
            disp = self.FILE.Get_byte_offset(offset)
            self.assertEqual(disp, offset)

    def testGetTypeExtent(self):
        extent = self.FILE.Get_type_extent(MPI.BYTE)
        self.assertEqual(extent, 1)

    def testGetErrhandler(self):
        eh = self.FILE.Get_errhandler()
        self.assertEqual(eh, MPI.ERRORS_RETURN)
        eh.Free()

    def testPyProps(self):
        file = self.FILE
        #
        self.assertEqual(file.size, 0)
        self.assertEqual(file.amode, self.amode)
        #
        group = file.group
        self.assertEqual(type(group), MPI.Group)
        self.assertEqual(file.group_size, group.Get_size())
        self.assertEqual(file.group_rank, group.Get_rank())
        group.Free()
        #
        info = file.info
        self.assertEqual(type(info), MPI.Info)
        file.info = info
        info.Free()
        #
        self.assertEqual(type(file.atomicity), bool)
        for atomicity in [True, False] * 4:
            file.atomicity = atomicity
            self.assertEqual(file.atomicity, atomicity)

    def testPickle(self):
        from pickle import dumps, loads

        with self.assertRaises(ValueError):
            loads(dumps(self.FILE))


class TestFileNull(unittest.TestCase):
    #
    def setUp(self):
        self.eh_save = MPI.FILE_NULL.Get_errhandler()

    def tearDown(self):
        MPI.FILE_NULL.Set_errhandler(self.eh_save)
        self.eh_save.Free()

    def testGetSetErrhandler(self):
        eh = MPI.FILE_NULL.Get_errhandler()
        self.assertEqual(eh, MPI.ERRORS_RETURN)
        eh.Free()
        MPI.FILE_NULL.Set_errhandler(MPI.ERRORS_ARE_FATAL)
        eh = MPI.FILE_NULL.Get_errhandler()
        self.assertEqual(eh, MPI.ERRORS_ARE_FATAL)
        eh.Free()
        MPI.FILE_NULL.Set_errhandler(MPI.ERRORS_RETURN)
        eh = MPI.FILE_NULL.Get_errhandler()
        self.assertEqual(eh, MPI.ERRORS_RETURN)
        eh.Free()


class TestFileSelf(BaseTestFile, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF
    prefix = BaseTestFile.prefix + (f"-{MPI.COMM_WORLD.Get_rank()}")


class TestFilePath(BaseTestFile, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF
    prefix = BaseTestFile.prefix + (f"-{MPI.COMM_WORLD.Get_rank()}")

    @staticmethod
    def _test_open_close(path):
        comm = MPI.COMM_SELF
        amode = MPI.MODE_CREATE | MPI.MODE_RDWR
        amode |= MPI.MODE_DELETE_ON_CLOSE
        try:
            fh = MPI.File.Open(comm, path, amode)
        except Exception:
            path.unlink()
            raise
        else:
            MPI.File.Close(fh)

    def testPath(self):
        name = maketemp(self.prefix)
        path = pathlib.Path(name)
        self._test_open_close(path)

    def testStr(self):
        name = maketemp(self.prefix)
        path = os.fsdecode(os.fsencode(name))
        self._test_open_close(path)

    def testBytes(self):
        name = maketemp(self.prefix)
        path = os.fsencode(name)
        self._test_open_close(path)


def have_feature():
    case = BaseTestFile()
    case.COMM = TestFileSelf.COMM
    case.prefix = TestFileSelf.prefix
    case.setUp()
    case.tearDown()


try:
    have_feature()
except NotImplementedError:
    unittest.disable(BaseTestFile, "mpi-file")
    unittest.disable(TestFileNull, "mpi-file")
    unittest.disable(TestFilePath, "mpi-file")


if __name__ == "__main__":
    unittest.main()



import os
import sys

import mpitestutil as testutil
import mpiunittest as unittest

from mpi4py import MPI

try:
    _ = sys.getrefcount
except AttributeError:

    class getrefcount:
        def __init__(self, arg):
            pass

        def __eq__(self, other):
            return True

        def __add__(self, other):
            return self

        def __sub__(self, other):
            return self


def memzero(m):
    try:
        m[:] = 0
    except IndexError:  # cffi buffer
        m[0 : len(m)] = b"\0" * len(m)


def ch3_sock():
    return "ch3:sock" in MPI.Get_library_version()


class TestWinNull(unittest.TestCase):
    #
    def testConstructor(self):
        win = MPI.Win()
        self.assertEqual(win, MPI.WIN_NULL)
        self.assertIsNot(win, MPI.WIN_NULL)

        def construct():
            MPI.Win((1, 2, 3))

        self.assertRaises(TypeError, construct)

    def testGetName(self):
        name = MPI.WIN_NULL.Get_name()
        self.assertEqual(name, "MPI_WIN_NULL")


class BaseTestWin:
    #
    COMM = MPI.COMM_NULL
    INFO = MPI.INFO_NULL
    CREATE_FLAVOR = MPI.UNDEFINED

    def testGetAttr(self):
        base = MPI.Get_address(self.memory)
        size = len(self.memory)
        unit = 1
        self.assertEqual(size, self.WIN.Get_attr(MPI.WIN_SIZE))
        self.assertEqual(unit, self.WIN.Get_attr(MPI.WIN_DISP_UNIT))
        self.assertEqual(base, self.WIN.Get_attr(MPI.WIN_BASE))

    def testMemory(self):
        memory = self.WIN.tomemory()
        self.assertEqual(memory.format, "B")
        pointer = MPI.Get_address(memory)
        length = len(memory)
        base, size, dunit = self.WIN.attrs
        self.assertEqual(size, length)
        self.assertEqual(dunit, 1)
        self.assertEqual(base, pointer)

    def testAttributes(self):
        base, size, unit = self.WIN.attrs
        self.assertEqual(base, MPI.Get_address(self.memory))
        self.assertEqual(size, len(self.memory))
        self.assertEqual(unit, 1)

    def testGetGroup(self):
        cgroup = self.COMM.Get_group()
        wgroup = self.WIN.Get_group()
        grpcmp = MPI.Group.Compare(cgroup, wgroup)
        cgroup.Free()
        wgroup.Free()
        self.assertEqual(grpcmp, MPI.IDENT)

    @unittest.skipMPI("impi(>=2021.15.0)")
    def testGetSetInfo(self):
        # info = MPI.INFO_NULL
        # self.WIN.Set_info(info)
        info = MPI.Info.Create()
        self.WIN.Set_info(info)
        info.Free()
        info = self.WIN.Get_info()
        self.WIN.Set_info(info)
        info.Free()

    def testGetSetErrhandler(self):
        for ERRHANDLER in [
            MPI.ERRORS_ARE_FATAL,
            MPI.ERRORS_RETURN,
            MPI.ERRORS_ARE_FATAL,
            MPI.ERRORS_RETURN,
        ]:
            errhdl_1 = self.WIN.Get_errhandler()
            self.assertNotEqual(errhdl_1, MPI.ERRHANDLER_NULL)
            self.WIN.Set_errhandler(ERRHANDLER)
            errhdl_2 = self.WIN.Get_errhandler()
            self.assertEqual(errhdl_2, ERRHANDLER)
            errhdl_2.Free()
            self.assertEqual(errhdl_2, MPI.ERRHANDLER_NULL)
            self.WIN.Set_errhandler(errhdl_1)
            errhdl_1.Free()
            self.assertEqual(errhdl_1, MPI.ERRHANDLER_NULL)

    def testGetSetName(self):
        try:
            name = self.WIN.Get_name()
            self.WIN.Set_name("mywin")
            self.assertEqual(self.WIN.Get_name(), "mywin")
            self.WIN.Set_name(name)
            self.assertEqual(self.WIN.Get_name(), name)
            self.WIN.name = self.WIN.name
        except NotImplementedError:
            self.skipTest("mpi-win-name")

    @unittest.skipIf(
        MPI.WIN_CREATE_FLAVOR == MPI.KEYVAL_INVALID, "mpi-win-flavor"
    )
    def testCreateFlavor(self):
        flavors = (
            MPI.WIN_FLAVOR_CREATE,
            MPI.WIN_FLAVOR_ALLOCATE,
            MPI.WIN_FLAVOR_DYNAMIC,
            MPI.WIN_FLAVOR_SHARED,
        )
        flavor = self.WIN.Get_attr(MPI.WIN_CREATE_FLAVOR)
        self.assertIn(flavor, flavors)
        self.assertEqual(flavor, self.WIN.flavor)
        self.assertEqual(flavor, self.CREATE_FLAVOR)

    @unittest.skipIf(MPI.WIN_MODEL == MPI.KEYVAL_INVALID, "mpi-win-model")
    def testMemoryModel(self):
        models = (MPI.WIN_SEPARATE, MPI.WIN_UNIFIED)
        model = self.WIN.Get_attr(MPI.WIN_MODEL)
        self.assertIn(model, models)
        self.assertEqual(model, self.WIN.model)

    def testSharedQuery(self):
        flavor = self.WIN.Get_attr(MPI.WIN_CREATE_FLAVOR)
        if flavor != MPI.WIN_FLAVOR_SHARED:
            if MPI.Get_version() < (4, 1):
                return
            if flavor == MPI.WIN_FLAVOR_DYNAMIC:
                return
            if unittest.is_mpi("impi"):
                i_mpi_compat = os.environ.get("I_MPI_COMPATIBILITY", "mpi-3.1")
                if i_mpi_compat in {"3", "4", "5", "mpi-3.1", "mpi-4.0"}:
                    return
                if flavor == MPI.WIN_FLAVOR_CREATE:  # mpi4py/mpi4py#665
                    return
            if unittest.is_mpi("mpich(<4.2.2)"):
                if flavor == MPI.WIN_FLAVOR_CREATE:
                    return
            if unittest.is_mpi("mpich"):  # pmodels/mpich#7499
                if "ch3:nemesis" in MPI.Get_library_version():
                    return
        memory = self.WIN.tomemory()
        address = MPI.Get_address(memory)
        length = len(memory)
        memories = self.COMM.allgather((address, length))
        for i in range(self.COMM.Get_size()):
            query = target = i
            mem, disp = self.WIN.Shared_query(query)
            base = MPI.Get_address(mem)
            size = len(mem)
            if flavor == MPI.WIN_FLAVOR_SHARED or size != 0:
                if self.COMM.Get_rank() == target:
                    self.assertEqual(base, memories[target][0])
                self.assertEqual(size, memories[target][1])
                self.assertEqual(disp, 1)
        query, target = MPI.PROC_NULL, 0
        mem, disp = self.WIN.Shared_query(query)
        base = MPI.Get_address(mem)
        size = len(mem)
        if flavor == MPI.WIN_FLAVOR_SHARED or size != 0:
            if self.COMM.Get_rank() == target:
                self.assertEqual(base, memories[target][0])
            self.assertEqual(size, memories[target][1])
            self.assertEqual(disp, 1)

    def testPyProps(self):
        win = self.WIN
        #
        group = win.group
        self.assertEqual(type(group), MPI.Group)
        self.assertEqual(win.group_size, group.Get_size())
        self.assertEqual(win.group_rank, group.Get_rank())
        group.Free()
        #
        if not unittest.is_mpi("impi(>=2021.15.0)"):
            info = win.info
            self.assertIs(type(info), MPI.Info)
            win.info = info
            info.Free()
        #
        self.assertEqual(type(win.attrs), tuple)
        self.assertEqual(type(win.flavor), int)
        self.assertEqual(type(win.model), int)
        self.assertEqual(type(win.name), str)
        win.name = "mywin"
        self.assertEqual(win.name, "mywin")

    def testPickle(self):
        from pickle import dumps, loads

        with self.assertRaises(ValueError):
            loads(dumps(self.WIN))


class BaseTestWinCreate(BaseTestWin):
    #
    CREATE_FLAVOR = MPI.WIN_FLAVOR_CREATE

    def setUp(self):
        blen = 10 + self.COMM.Get_rank()
        self.memory = MPI.Alloc_mem(blen)
        memzero(self.memory)
        self.WIN = MPI.Win.Create(self.memory, 1, self.INFO, self.COMM)

    def tearDown(self):
        self.WIN.Free()
        MPI.Free_mem(self.memory)


class BaseTestWinAllocate(BaseTestWin):
    #
    CREATE_FLAVOR = MPI.WIN_FLAVOR_ALLOCATE

    def setUp(self):
        blen = 10 + self.COMM.Get_rank()
        self.WIN = MPI.Win.Allocate(blen, 1, self.INFO, self.COMM)
        self.memory = self.WIN.tomemory()
        memzero(self.memory)

    def tearDown(self):
        self.WIN.Free()


class BaseTestWinAllocateShared(BaseTestWin):
    #
    CREATE_FLAVOR = MPI.WIN_FLAVOR_SHARED

    def setUp(self):
        if self.COMM.Get_size() > 1:
            try:
                self.COMM = self.COMM.Split_type(MPI.COMM_TYPE_SHARED)
            except NotImplementedError:
                self.skipTest("mpi-comm-split_type")
        else:
            self.COMM = self.COMM.Dup()
        try:
            blen = 10 + self.COMM.Get_rank()
            self.WIN = MPI.Win.Allocate_shared(blen, 1, self.INFO, self.COMM)
        except Exception:
            self.COMM.Free()
            raise
        self.memory = self.WIN.tomemory()
        memzero(self.memory)

    def tearDown(self):
        self.COMM.Free()
        self.WIN.Free()

    @unittest.skipMPI("mpich(>=4.2.0,<4.3.0)", sys.platform == "linux")
    def testSharedQuery(self):
        super().testSharedQuery()


@unittest.skipMPI("impi(>=2021.14.0,<2021.15.0)", testutil.github())
class BaseTestWinCreateDynamic(BaseTestWin):
    #
    CREATE_FLAVOR = MPI.WIN_FLAVOR_DYNAMIC

    def setUp(self):
        self.WIN = MPI.Win.Create_dynamic(self.INFO, self.COMM)

    def tearDown(self):
        self.WIN.Free()

    def testGetAttr(self):
        base = self.WIN.Get_attr(MPI.WIN_BASE)
        size = self.WIN.Get_attr(MPI.WIN_SIZE)
        self.assertEqual(base, 0)
        self.assertEqual(size, 0)

    def testMemory(self):
        memory = self.WIN.tomemory()
        self.assertEqual(memory.format, "B")
        base = MPI.Get_address(memory)
        size = len(memory)
        self.assertEqual(base, 0)
        self.assertEqual(size, 0)

    def testAttributes(self):
        base, size, _ = self.WIN.attrs
        self.assertEqual(base, 0)
        self.assertEqual(size, 0)

    @unittest.skipMPI("msmpi(<9.1.0)")
    def testAttachDetach(self):
        mem1 = MPI.Alloc_mem(8)
        mem2 = MPI.Alloc_mem(16)
        mem3 = MPI.Alloc_mem(32)
        for mem in (mem1, mem2, mem3):
            self.WIN.Attach(mem)
            self.testMemory()
            self.WIN.Detach(mem)
        for mem in (mem1, mem2, mem3):
            self.WIN.Attach(mem)
        self.testMemory()
        for mem in (mem1, mem2, mem3):
            self.WIN.Detach(mem)
        for mem in (mem1, mem2, mem3):
            self.WIN.Attach(mem)
        self.testMemory()
        for mem in (mem3, mem2, mem1):
            self.WIN.Detach(mem)
        MPI.Free_mem(mem1)
        MPI.Free_mem(mem2)
        MPI.Free_mem(mem3)


class TestWinCreateSelf(BaseTestWinCreate, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


@unittest.skipMPI("openmpi(<1.4.0)")
class TestWinCreateWorld(BaseTestWinCreate, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


class TestWinAllocateSelf(BaseTestWinAllocate, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


@unittest.skipMPI("openmpi(<1.4.0)")
class TestWinAllocateWorld(BaseTestWinAllocate, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


class TestWinAllocateSharedSelf(BaseTestWinAllocateShared, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


@unittest.skipMPI("mpich", ch3_sock() and MPI.COMM_WORLD.Get_size() > 1)
class TestWinAllocateSharedWorld(BaseTestWinAllocateShared, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


class TestWinCreateDynamicSelf(BaseTestWinCreateDynamic, unittest.TestCase):
    #
    COMM = MPI.COMM_SELF


class TestWinCreateDynamicWorld(BaseTestWinCreateDynamic, unittest.TestCase):
    #
    COMM = MPI.COMM_WORLD


try:
    MPI.Win.Create(MPI.BOTTOM, 1, MPI.INFO_NULL, MPI.COMM_SELF).Free()
except (NotImplementedError, MPI.Exception):
    unittest.disable(BaseTestWinCreate, "mpi-win-create")
try:
    MPI.Win.Allocate(1, 1, MPI.INFO_NULL, MPI.COMM_SELF).Free()
except (NotImplementedError, MPI.Exception):
    unittest.disable(BaseTestWinAllocate, "mpi-win-allocate")
try:
    MPI.Win.Allocate_shared(1, 1, MPI.INFO_NULL, MPI.COMM_SELF).Free()
except (NotImplementedError, MPI.Exception):
    unittest.disable(BaseTestWinAllocateShared, "mpi-win-shared")
try:
    MPI.Win.Create_dynamic(MPI.INFO_NULL, MPI.COMM_SELF).Free()
except (NotImplementedError, MPI.Exception):
    unittest.disable(BaseTestWinCreateDynamic, "mpi-win-dynamic")


if __name__ == "__main__":
    unittest.main()



import numpy as np

from mpi4py import MPI

tic = MPI.Wtime()

x1 = -2.0
x2 = +1.0
y1 = -1.0
y2 = +1.0

w = 150
h = 100
maxit = 127


def mandelbrot(x, y, maxit):
    c = x + y * 1j
    z = 0 + 0j
    it = 0
    while abs(z) < 2 and it < maxit:
        z = z**2 + c
        it += 1
    return it


comm = MPI.COMM_WORLD
size = comm.Get_size()
rank = comm.Get_rank()

rmsg = np.empty(4, dtype="f")
imsg = np.empty(3, dtype="i")

if rank == 0:
    rmsg[:] = [x1, x2, y1, y2]
    imsg[:] = [w, h, maxit]

comm.Bcast([rmsg, MPI.FLOAT], root=0)
comm.Bcast([imsg, MPI.INT], root=0)

x1, x2, y1, y2 = (float(r) for r in rmsg)
w, h, maxit = (int(i) for i in imsg)
dx = (x2 - x1) / w
dy = (y2 - y1) / h

# number of lines to compute here
N = h // size + (h % size > rank)
N = np.array(N, dtype="i")
# indices of lines to compute here
I = np.arange(rank, h, size, dtype="i")  # noqa: E741
# compute local lines
C = np.empty([N, w], dtype="i")
for k in np.arange(N):
    y = y1 + I[k] * dy
    for j in np.arange(w):
        x = x1 + j * dx
        C[k, j] = mandelbrot(x, y, maxit)
# gather results at root
counts = 0
indices = None
cdata = None
if rank == 0:
    counts = np.empty(size, dtype="i")
    indices = np.empty(h, dtype="i")
    cdata = np.empty([h, w], dtype="i")
comm.Gather(sendbuf=[N, MPI.INT], recvbuf=[counts, MPI.INT], root=0)
comm.Gatherv(
    sendbuf=[I, MPI.INT], recvbuf=[indices, (counts, None), MPI.INT], root=0
)
comm.Gatherv(
    sendbuf=[C, MPI.INT], recvbuf=[cdata, (counts * w, None), MPI.INT], root=0
)
# reconstruct full result at root
if rank == 0:
    M = np.zeros([h, w], dtype="i")
    M[indices, :] = cdata

toc = MPI.Wtime()
wct = comm.gather(toc - tic, root=0)
if rank == 0:
    for task, time in enumerate(wct):
        print(f"wall clock time: {time:8.2f} seconds (task {task:d})")

    def mean(seq):
        return sum(seq) / len(seq)

    print(f"all tasks, mean: {mean(wct):8.2f} seconds")
    print(f"all tasks, min:  {min(wct):8.2f} seconds")
    print(f"all tasks, max:  {max(wct):8.2f} seconds")
    print(f"all tasks, sum:  {sum(wct):8.2f} seconds")

# eye candy (requires matplotlib)
if rank == 0:
    import contextlib

    with contextlib.suppress(Exception):
        from matplotlib import pyplot as plt

        plt.imshow(M, aspect="equal")
        try:
            plt.nipy_spectral()
        except AttributeError:
            plt.spectral()
        plt.pause(2)

MPI.COMM_WORLD.Barrier()



import mpitestutil as testutil
import mpiunittest as unittest

from mpi4py import MPI


class TestMPIABI(unittest.TestCase):
    #
    def testGetVersion(self):
        version = MPI.Get_abi_version()
        self.assertTrue(version >= (1, 0) or version == (-1, -1))

    def testGetInfo(self):
        info = MPI.Get_abi_info()
        for typename in (
            "aint",
            "count",
            "offset",
        ):
            with self.subTest(typename=typename):
                key = f"mpi_{typename}_size"
                sizestr = info.Get(key)
                self.assertIsNotNone(sizestr)
                datatype = getattr(MPI, typename.upper())
                if testutil.has_datatype(datatype):
                    self.assertEqual(datatype.Get_size(), int(sizestr))
        info.Free()

    def testGetFortranInfo(self):
        info = MPI.Get_abi_fortran_info()
        if info == MPI.INFO_NULL:
            return
        for typename in (
            "logical",
            "integer",
            "real",
            "double_precision",
        ):
            with self.subTest(typename=typename):
                key = f"mpi_{typename}_size"
                sizestr = info.Get(key)
                self.assertIsNotNone(sizestr)
                datatype = getattr(MPI, typename.upper())
                if testutil.has_datatype(datatype):
                    self.assertEqual(datatype.Get_size(), int(sizestr))
        str2bool = {"true": True, "false": False}
        for typeclass, typesizes in (
            ("logical", (1, 2, 4, 8, 16)),
            ("integer", (1, 2, 4, 8, 16)),
            ("real", (2, 4, 8, 16)),
            ("complex", (4, 8, 16, 32)),
            ("double_complex", ("",)),
        ):
            for typesize in typesizes:
                typename = f"{typeclass}{typesize}"
                datatype = getattr(MPI, typename.upper())
                with self.subTest(typename=typename):
                    key = f"mpi_{typename}_supported"
                    supported = info.Get(key)
                    self.assertIn(supported, ("true", "false"))
                    self.assertEqual(
                        testutil.has_datatype(datatype),
                        str2bool[supported],
                    )
        info.Free()


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
"""Run Cython with custom options."""

import os
import pathlib
import sys

from Cython.Compiler.Main import main as cython_main


def cythonize(args=None):
    """Run `cython --3str --cleanup 3 <args>...`."""
    if args is None:
        argv = sys.argv[:]
    else:
        arg0 = str(pathlib.Path(__file__).resolve())
        argv = [arg0, *args]

    if "--cleanup" not in argv:
        argv[1:1] = ["--cleanup", "3"]
    if "--3str" not in argv:
        argv[1:1] = ["--3str"]

    cwd = pathlib.Path.cwd()
    sys_argv = sys.argv[:]
    try:
        sys.argv[:] = argv
        cython_main(command_line=1)
    except SystemExit as exc:
        return exc.code
    else:
        return 0
    finally:
        os.chdir(cwd)
        sys.argv[:] = sys_argv


def main():
    """Entry-point to run Cython with custom options."""
    args = sys.argv[1:]
    if not args:
        topdir = pathlib.Path(__file__).resolve().parent.parent
        source = pathlib.Path() / "src" / "mpi4py" / "MPI.pyx"
        target = pathlib.Path() / "src" / "mpi4py" / "MPI.c"
        args += ["--working", str(topdir)]
        args += [str(source), "--output-file", str(target)]
    sys.exit(cythonize(args))


if __name__ == "__main__":
    main()



cdef class Op:
    """
    Reduction operation.
    """

    def __cinit__(self, Op op: Op | None = None):
        cinit(self, op)

    def __dealloc__(self):
        dealloc(self)

    def __richcmp__(self, other, int op):
        if not isinstance(other, Op): return NotImplemented
        return richcmp(self, other, op)

    def __bool__(self) -> bool:
        return nonnull(self)

    def __reduce__(self) -> str | tuple[Any, ...]:
        return reduce_Op(self)

    def __call__(self, x: Any, y: Any) -> Any:
        cdef int index = op_user_id_get(self)
        return op_call(self.ob_mpi, index, x, y)

    property handle:
        """MPI handle."""
        def __get__(self) -> int:
            return tohandle(self)

    @classmethod
    def fromhandle(cls, handle: int) -> Op:
        """
        Create object from MPI handle.
        """
        return fromhandle(<MPI_Op> <Py_uintptr_t> handle)

    def free(self) -> None:
        """
        Call `Free` if not null or predefined.
        """
        safefree(self)

    # User-Defined Reduction Operations
    # ---------------------------------

    @classmethod
    def Create(
        cls,
        function: Callable[[Buffer, Buffer, Datatype], None],
        bint commute: bool = False,
    ) -> Self:
        """
        Create a user-defined reduction operation.
        """
        cdef Op self = <Op>New(cls)
        cdef MPI_User_function   *fn_i = NULL
        cdef MPI_User_function_c *fn_c = NULL
        op_user_new(self, function, &fn_i, &fn_c)
        try:
            try:
                CHKERR( MPI_Op_create_c(fn_c, commute, &self.ob_mpi) )
            except NotImplementedError:                               # ~> legacy
                CHKERR( MPI_Op_create(fn_i, commute, &self.ob_mpi) )  # ~> legacy
        except:                # ~> uncovered  # noqa
            op_user_del(self)  # ~> uncovered
            raise              # ~> uncovered
        return self

    def Free(self) -> None:
        """
        Free a user-defined reduction operation.
        """
        cdef MPI_Op save = self.ob_mpi
        CHKERR( MPI_Op_free(&self.ob_mpi) )
        if constobj(self): self.ob_mpi = save
        op_user_del(self)

    # Process-local reduction
    # -----------------------

    def Is_commutative(self) -> bool:
        """
        Query reduction operations for their commutativity.
        """
        cdef int flag = 0
        CHKERR( MPI_Op_commutative(self.ob_mpi, &flag) )
        return <bint>flag

    property is_commutative:
        """Is a commutative operation."""
        def __get__(self) -> bool:
            return self.Is_commutative()

    def Reduce_local(self, inbuf: BufSpec, inoutbuf: BufSpec) -> None:
        """
        Apply a reduction operation to local data.
        """
        # get *in* and *inout* buffers
        cdef _p_msg_cco m = message_cco()
        m.for_cro_recv(inoutbuf, 0)
        m.for_cro_send(inbuf, 0)
        m.chk_cro_args()
        # do local reduction
        with nogil: CHKERR( MPI_Reduce_local_c(
            m.sbuf, m.rbuf, m.rcount, m.rtype, self.ob_mpi) )

    property is_predefined:
        """Is a predefined operation."""
        def __get__(self) -> bool:
            cdef MPI_Op op = self.ob_mpi
            return (
                op == MPI_OP_NULL or
                op == MPI_MAX or
                op == MPI_MIN or
                op == MPI_SUM or
                op == MPI_PROD or
                op == MPI_LAND or
                op == MPI_BAND or
                op == MPI_LOR or
                op == MPI_BOR or
                op == MPI_LXOR or
                op == MPI_BXOR or
                op == MPI_MAXLOC or
                op == MPI_MINLOC or
                op == MPI_REPLACE or
                op == MPI_NO_OP
            )

    # Integer Handle
    # --------------

    def toint(self) -> int:
        """
        Translate object to integer handle.
        """
        return MPI_Op_toint(self.ob_mpi)

    @classmethod
    def fromint(cls, arg: int, /) -> Op:
        """
        Translate integer handle to object.
        """
        return fromhandle(MPI_Op_fromint(arg))

    # Fortran Handle
    # --------------

    def py2f(self) -> int:
        """
        """
        return MPI_Op_c2f(self.ob_mpi)

    @classmethod
    def f2py(cls, arg: int) -> Op:
        """
        """
        return fromhandle(MPI_Op_f2c(arg))


cdef Op __OP_NULL__ = def_Op( MPI_OP_NULL , "OP_NULL" )
cdef Op __MAX__     = def_Op( MPI_MAX     , "MAX"     )
cdef Op __MIN__     = def_Op( MPI_MIN     , "MIN"     )
cdef Op __SUM__     = def_Op( MPI_SUM     , "SUM"     )
cdef Op __PROD__    = def_Op( MPI_PROD    , "PROD"    )
cdef Op __LAND__    = def_Op( MPI_LAND    , "LAND"    )
cdef Op __BAND__    = def_Op( MPI_BAND    , "BAND"    )
cdef Op __LOR__     = def_Op( MPI_LOR     , "LOR"     )
cdef Op __BOR__     = def_Op( MPI_BOR     , "BOR"     )
cdef Op __LXOR__    = def_Op( MPI_LXOR    , "LXOR"    )
cdef Op __BXOR__    = def_Op( MPI_BXOR    , "BXOR"    )
cdef Op __MAXLOC__  = def_Op( MPI_MAXLOC  , "MAXLOC"  )
cdef Op __MINLOC__  = def_Op( MPI_MINLOC  , "MINLOC"  )
cdef Op __REPLACE__ = def_Op( MPI_REPLACE , "REPLACE" )
cdef Op __NO_OP__   = def_Op( MPI_NO_OP   , "NO_OP"   )


# Predefined operation handles
# ----------------------------

OP_NULL = __OP_NULL__  #: Null
MAX     = __MAX__      #: Maximum
MIN     = __MIN__      #: Minimum
SUM     = __SUM__      #: Sum
PROD    = __PROD__     #: Product
LAND    = __LAND__     #: Logical and
BAND    = __BAND__     #: Bit-wise and
LOR     = __LOR__      #: Logical or
BOR     = __BOR__      #: Bit-wise or
LXOR    = __LXOR__     #: Logical xor
BXOR    = __BXOR__     #: Bit-wise xor
MAXLOC  = __MAXLOC__   #: Maximum and location
MINLOC  = __MINLOC__   #: Minimum and location
REPLACE = __REPLACE__  #: Replace (for RMA)
NO_OP   = __NO_OP__    #: No-op   (for RMA)



# -----------------------------------------------------------------------------


@cython.final
@cython.internal
cdef class _p_rs:

    cdef int         count
    cdef MPI_Request *requests
    cdef MPI_Status  *status, _status[1]
    cdef MPI_Status  *statuses
    cdef int         outcount
    cdef int         *indices
    cdef object      arg_req
    cdef object      buf_req
    cdef object      buf_sts
    cdef object      buf_ids
    cdef MPI_Status  tmp_sts

    def __cinit__(self):
        self.count    = 0
        self.requests = NULL
        self.status   = MPI_STATUS_IGNORE
        self.statuses = MPI_STATUSES_IGNORE
        self.outcount = MPI_UNDEFINED
        self.indices  = NULL
        self.arg_req = None
        self.buf_req = None
        self.buf_sts = None
        self.buf_ids = None

    cdef int set_request(self, Request request) except -1:
        self.arg_req = request
        return 0

    cdef int set_status(self, Status status) except -1:
        if status is not None:
            self.status = &status.ob_mpi
        else:
            self.status = &self.tmp_sts
            <void>MPI_Status_set_source ( self.status, MPI_ANY_SOURCE )
            <void>MPI_Status_set_tag    ( self.status, MPI_ANY_TAG    )
            <void>MPI_Status_set_error  ( self.status, MPI_SUCCESS    )
        return 0

    cdef int set_requests(self, requests) except -1:
        self.arg_req = requests
        cdef Py_ssize_t count = len(requests)
        self.count = <int>count
        self.outcount = <int>count
        self.buf_req = allocate(self.count, sizeof(MPI_Request), &self.requests)
        for i in range(self.count):
            self.requests[i] = (<Request?>requests[i]).ob_mpi
        return 0

    cdef int add_statuses(self) except -1:
        cdef MPI_Status *status = &self.tmp_sts
        <void>MPI_Status_set_source ( status, MPI_ANY_SOURCE )
        <void>MPI_Status_set_tag    ( status, MPI_ANY_TAG    )
        <void>MPI_Status_set_error  ( status, MPI_SUCCESS    )
        self.buf_sts = allocate(self.count, sizeof(MPI_Status), &self.statuses)
        for i in range(self.count):
            self.statuses[i] = status[0]
        return 0

    cdef int add_indices(self) except -1:
        self.outcount = MPI_UNDEFINED
        self.buf_ids = newarray(self.count, &self.indices)
        return 0

    cdef int acquire(self, requests, statuses=None) except -1:
        self.set_requests(requests)
        if statuses is not None:
            self.add_statuses()
        return 0

    cdef int release(self, statuses=None) except -1:
        cdef Request request
        cdef Py_ssize_t nr = self.count
        for i in range(nr):
            request = <Request>self.arg_req[i]
            request.ob_mpi = self.requests[i]
            if request.ob_mpi == MPI_REQUEST_NULL:
                if request.ob_buf is not None:
                    request.ob_buf = None
        #
        if statuses is None:
            return 0
        if self.outcount == MPI_UNDEFINED:
            return 0
        cdef Py_ssize_t outcount = self.outcount
        cdef Py_ssize_t ns = len(statuses)
        if outcount > ns:
            if isinstance(statuses, list):
                statuses += [
                    <Status>New(Status)
                    for _ in range (ns, outcount)
                ]
                ns = outcount
        for i in range(min(nr, ns)):
            (<Status?>statuses[i]).ob_mpi = self.statuses[i]
        return 0

    cdef object get_buffer(self, Py_ssize_t index):
        cdef Request request
        if index >= 0:
            if self.indices != NULL:
                index = self.indices[index]
            request = <Request>self.arg_req[index]
        else:
            request = <Request>self.arg_req
        cdef object buf = request.ob_buf
        if request.ob_mpi == MPI_REQUEST_NULL:
            if request.ob_buf is not None:
                request.ob_buf = None
        return buf

    cdef object get_result(self):
        return self.get_object(-1)

    cdef object get_object(self, Py_ssize_t index):
        return PyMPI_load(self.get_buffer(index), self.status)

    cdef object get_objects(self):
        if self.outcount == MPI_UNDEFINED: return None
        return [
            PyMPI_load(self.get_buffer(i), &self.statuses[i])
            for i in range(self.outcount)
        ]

    cdef object get_indices(self):
        if self.outcount == MPI_UNDEFINED: return None
        return [self.indices[i] for i in range(self.outcount)]

# -----------------------------------------------------------------------------


@cython.final
@cython.internal
cdef class _p_greq:

    cdef object query_fn
    cdef object free_fn
    cdef object cancel_fn
    cdef tuple args
    cdef dict  kwargs

    def __cinit__(self, query_fn, free_fn, cancel_fn, args, kwargs):
        self.query_fn  = query_fn
        self.free_fn   = free_fn
        self.cancel_fn = cancel_fn
        self.args   = tuple(args)  if args   is not None else ()
        self.kwargs = dict(kwargs) if kwargs is not None else {}

    cdef int query(self, MPI_Status *status) except -1:
        <void>MPI_Status_set_source(status, MPI_ANY_SOURCE)
        <void>MPI_Status_set_tag(status, MPI_ANY_TAG)
        <void>MPI_Status_set_error(status, MPI_SUCCESS)
        <void>MPI_Status_set_elements_c(status, MPI_BYTE, 0)
        <void>MPI_Status_set_cancelled(status, 0)
        cdef Status sts
        if self.query_fn is not None:
            sts = <Status>New(Status)
            sts.ob_mpi = status[0]
            self.query_fn(sts, *self.args, **self.kwargs)
            status[0] = sts.ob_mpi
            if self.cancel_fn is None:
                <void>MPI_Status_set_cancelled(status, 0)
        return MPI_SUCCESS

    cdef int free(self) except -1:
        if self.free_fn is not None:
            self.free_fn(*self.args, **self.kwargs)
        return MPI_SUCCESS

    cdef int cancel(self, bint completed) except -1:
        if self.cancel_fn is not None:
            self.cancel_fn(completed, *self.args, **self.kwargs)
        return MPI_SUCCESS


cdef int greq_query(
    void *extra_state,
    MPI_Status *status,
) noexcept with gil:
    cdef _p_greq state = <_p_greq>extra_state
    cdef int ierr = MPI_SUCCESS
    cdef object exc
    try:
        state.query(status)
    except BaseException as exc:
        ierr = PyMPI_HandleException(exc)
    return ierr


cdef int greq_free(
    void *extra_state,
) noexcept with gil:
    cdef _p_greq state = <_p_greq>extra_state
    cdef int ierr = MPI_SUCCESS
    cdef object exc
    try:
        state.free()
    except BaseException as exc:
        ierr = PyMPI_HandleException(exc)
    finally:
        Py_DECREF(<object>extra_state)
    return ierr


cdef int greq_cancel(
    void *extra_state,
    int completed,
) noexcept with gil:
    cdef _p_greq state = <_p_greq>extra_state
    cdef int ierr = MPI_SUCCESS
    cdef object exc
    try:
        state.cancel(completed)
    except BaseException as exc:
        ierr = PyMPI_HandleException(exc)
    return ierr


@cython.callspec("MPIAPI")
cdef int greq_query_fn(
    void *extra_state,
    MPI_Status *status,
) noexcept nogil:
    if extra_state == NULL:    return MPI_ERR_INTERN
    if status == NULL:         return MPI_ERR_INTERN
    if not Py_IsInitialized(): return MPI_ERR_INTERN
    if not py_module_alive():  return MPI_ERR_INTERN
    return greq_query(extra_state, status)


@cython.callspec("MPIAPI")
cdef int greq_free_fn(
    void *extra_state,
) noexcept nogil:
    if extra_state == NULL:    return MPI_ERR_INTERN
    if not Py_IsInitialized(): return MPI_ERR_INTERN
    if not py_module_alive():  return MPI_ERR_INTERN
    return greq_free(extra_state)


@cython.callspec("MPIAPI")
cdef int greq_cancel_fn(
    void *extra_state,
    int completed,
) noexcept nogil:
    if extra_state == NULL:    return MPI_ERR_INTERN
    if not Py_IsInitialized(): return MPI_ERR_INTERN
    if not py_module_alive():  return MPI_ERR_INTERN
    return greq_cancel(extra_state, completed)


# -----------------------------------------------------------------------------



# Author:  Lisandro Dalcin
# Contact: dalcinl@gmail.com

# --

from mpi4py.libmpi cimport MPI_Aint
from mpi4py.libmpi cimport MPI_Offset
from mpi4py.libmpi cimport MPI_Count
from mpi4py.libmpi cimport MPI_Status
from mpi4py.libmpi cimport MPI_Datatype
from mpi4py.libmpi cimport MPI_Request
from mpi4py.libmpi cimport MPI_Message
from mpi4py.libmpi cimport MPI_Op
from mpi4py.libmpi cimport MPI_Group
from mpi4py.libmpi cimport MPI_Info
from mpi4py.libmpi cimport MPI_Errhandler
from mpi4py.libmpi cimport MPI_Session
from mpi4py.libmpi cimport MPI_Comm
from mpi4py.libmpi cimport MPI_Win
from mpi4py.libmpi cimport MPI_File

# --

cdef extern from *:
    """
    #if (MPI_VERSION < 3) && !defined(PyMPI_HAVE_MPI_Message)
    #if !defined(MPI_Message) && !defined(MPI_MESSAGE_NULL)
    typedef void *PyMPI_MPI_Message;
    #define MPI_Message PyMPI_MPI_Message
    #endif
    #endif

    #if (MPI_VERSION < 4) && !defined(PyMPI_HAVE_MPI_Session)
    #if !defined(MPI_Session) && !defined(MPI_SESSION_NULL)
    typedef void *PyMPI_MPI_Session;
    #define MPI_Session PyMPI_MPI_Session
    #endif
    #endif
    """
    ctypedef MPI_Aint   Aint   "MPI_Aint"
    ctypedef MPI_Offset Offset "MPI_Offset"
    ctypedef MPI_Count  Count  "MPI_Count"

ctypedef api class Datatype [
    type   PyMPIDatatype_Type,
    object PyMPIDatatypeObject,
]:
    cdef MPI_Datatype ob_mpi
    cdef unsigned     flags
    cdef object     __weakref__

ctypedef api class Status [
    type   PyMPIStatus_Type,
    object PyMPIStatusObject,
]:
    cdef MPI_Status ob_mpi
    cdef unsigned   flags
    cdef object   __weakref__

ctypedef api class Request [
    type   PyMPIRequest_Type,
    object PyMPIRequestObject,
]:
    cdef MPI_Request ob_mpi
    cdef unsigned    flags
    cdef object    __weakref__
    cdef object      ob_buf

ctypedef api class Prequest(Request) [
    type   PyMPIPrequest_Type,
    object PyMPIPrequestObject,
]:
    pass

ctypedef api class Grequest(Request) [
    type   PyMPIGrequest_Type,
    object PyMPIGrequestObject,
]:
    cdef MPI_Request ob_grequest

ctypedef api class Message [
    type   PyMPIMessage_Type,
    object PyMPIMessageObject,
]:
    cdef MPI_Message ob_mpi
    cdef unsigned    flags
    cdef object    __weakref__
    cdef object      ob_buf

ctypedef api class Op [
    type   PyMPIOp_Type,
    object PyMPIOpObject,
]:
    cdef MPI_Op   ob_mpi
    cdef unsigned flags
    cdef object __weakref__

ctypedef api class Group [
    type   PyMPIGroup_Type,
    object PyMPIGroupObject,
]:
    cdef MPI_Group ob_mpi
    cdef unsigned  flags
    cdef object  __weakref__

ctypedef api class Info [
    type   PyMPIInfo_Type,
    object PyMPIInfoObject,
]:
    cdef MPI_Info ob_mpi
    cdef unsigned flags
    cdef object __weakref__

ctypedef api class Errhandler [
    type   PyMPIErrhandler_Type,
    object PyMPIErrhandlerObject,
]:
    cdef MPI_Errhandler ob_mpi
    cdef unsigned       flags
    cdef object       __weakref__

ctypedef api class Session [
    type   PyMPISession_Type,
    object PyMPISessionObject,
]:
    cdef MPI_Session ob_mpi
    cdef unsigned    flags
    cdef object     __weakref__

ctypedef api class Comm [
    type   PyMPIComm_Type,
    object PyMPICommObject,
]:
    cdef MPI_Comm ob_mpi
    cdef unsigned flags
    cdef object __weakref__

ctypedef api class Intracomm(Comm) [
    type   PyMPIIntracomm_Type,
    object PyMPIIntracommObject,
]:
    pass

ctypedef api class Topocomm(Intracomm) [
    type   PyMPITopocomm_Type,
    object PyMPITopocommObject,
]:
    pass

ctypedef api class Cartcomm(Topocomm) [
    type   PyMPICartcomm_Type,
    object PyMPICartcommObject,
]:
    pass

ctypedef api class Graphcomm(Topocomm) [
    type   PyMPIGraphcomm_Type,
    object PyMPIGraphcommObject,
]:
    pass

ctypedef api class Distgraphcomm(Topocomm) [
    type   PyMPIDistgraphcomm_Type,
    object PyMPIDistgraphcommObject,
]:
    pass

ctypedef api class Intercomm(Comm) [
    type   PyMPIIntercomm_Type,
    object PyMPIIntercommObject,
]:
    pass

ctypedef api class Win [
    type   PyMPIWin_Type,
    object PyMPIWinObject,
]:
    cdef MPI_Win  ob_mpi
    cdef unsigned flags
    cdef object __weakref__
    cdef object   ob_mem

ctypedef api class File [
    type   PyMPIFile_Type,
    object PyMPIFileObject,
]:
    cdef MPI_File ob_mpi
    cdef unsigned flags
    cdef object __weakref__

# --



# -----------------------------------------------------------------------------

cdef extern from "Python.h":
    int    PyUnicode_Check(object)
    object PyUnicode_AsUTF8String(object)
    object PyUnicode_FromString(const char[])
    object PyUnicode_FromStringAndSize(const char[], Py_ssize_t)
    object PyBytes_FromString(const char[])
    object PyBytes_FromStringAndSize(const char[], Py_ssize_t)
    int    PyBytes_AsStringAndSize(object, char*[], Py_ssize_t*) except -1

# -----------------------------------------------------------------------------

cdef inline object asmpistr(object ob, char *s[]):
    if PyUnicode_Check(ob):
        ob = PyUnicode_AsUTF8String(ob)
    PyBytes_AsStringAndSize(ob, s, NULL)
    return ob

cdef inline object tompistr(const char s[], int n):
    return PyUnicode_FromStringAndSize(s, n)

cdef inline object mpistr(const char s[]):
    return PyUnicode_FromString(s)

cdef inline object pystr(const char s[]):
    return PyUnicode_FromString(s)

# -----------------------------------------------------------------------------

cdef extern from * nogil:
    """
    static int PyMPI_tolower(int c)
    {
      return (c >= 'A' && c <= 'Z') ? c + ('a' - 'A') : c;
    }
    static int PyMPI_strcasecmp(const char *s1, const char *s2)
    {
      int c1, c2;
      do {
         c1 = PyMPI_tolower((int)*s1++);
         c2 = PyMPI_tolower((int)*s2++);
      } while (c1 && c1 == c2);
      return c1 - c2;
    }
    """
    int PyMPI_strcasecmp(const char[], const char[])

cdef inline int cstr2bool(const char s[]) noexcept nogil:
    cdef const char **T = [b"true",  b"yes", b"on",  b"y", b"1"], *t = NULL
    cdef const char **F = [b"false", b"no",  b"off", b"n", b"0"], *f = NULL
    if s == NULL: return 0
    if s[0] == 0: return 0
    for f in F[:5]:
        if PyMPI_strcasecmp(s, f) == 0:
            return 0
    for t in T[:5]:
        if PyMPI_strcasecmp(s, t) == 0:
            return 1
    return -1

cdef inline int cstr_is_bool(const char s[]) noexcept nogil:
    return cstr2bool(s) >= 0

cdef inline int cstr_is_uint(const char s[]) noexcept nogil:
    if s == NULL: return 0
    if s[0] == 0: return 0
    for i in range(<Py_ssize_t>64):
        if s[i] == c' ' or s[i] == c'\t': continue
        if s[i] >= c'0' and s[i] <= c'9': continue
        if s[i] == 0: return 1
        break
    return 0

# -----------------------------------------------------------------------------



import mpiunittest as unittest

from mpi4py import MPI

try:
    import pickle as pyPickle
except ImportError:
    pyPickle = None

try:
    import dill
except ImportError:
    dill = None

try:
    import marshal
except ImportError:
    marshal = None

try:
    import json
except ImportError:
    json = None

try:
    import yaml

    yaml.dump(None)
except ImportError:
    yaml = None

OBJS = [
    None,
    True,
    False,
    7,
    1 << 32,
    1.62,
    1 + 2j,
    "qwerty",
    (0, 1, 2),
    [0, 1, 2],
    {"a": 0, "b": 1},
]


def tobytes(s):
    return memoryview(s).tobytes()


class TestPickle(unittest.TestCase):
    #
    def setUp(self):
        self.pickle = MPI.pickle

    def tearDown(self):
        self.pickle.__init__()

    def do_pickle(self, obj, pickle):
        comm = MPI.COMM_SELF
        o = comm.sendrecv(obj, 0, 0, None, 0, 0)
        self.assertEqual(obj, o)
        s = pickle.dumps(obj)
        o = pickle.loads(s)
        self.assertEqual(obj, o)

    def testDefault(self):
        pickle = self.pickle
        protocols = [0, 1, 2, 3, 4, 5]
        protocols.extend((-1, None))
        for proto in protocols:
            pickle.__init__(protocol=proto)
            for obj in OBJS:
                self.do_pickle(obj, pickle)
            self.do_pickle(OBJS, pickle)

    def testPyPickle(self):
        pickle = self.pickle
        dumps = pyPickle.dumps
        loads = pyPickle.loads
        protocols = [0, 1, 2, 3, 4, 5]
        protocols.extend((-1, None))
        for proto in protocols:
            pickle.__init__(dumps, loads, proto)
            for obj in OBJS:
                self.do_pickle(obj, pickle)
            self.do_pickle(OBJS, pickle)

    @unittest.skipIf(dill is None, "dill")
    def testDill(self):
        pickle = self.pickle
        dumps = dill.dumps
        loads = dill.loads
        protocols = list(range(dill.HIGHEST_PROTOCOL + 1))
        protocols.extend((-1, None))
        for proto in protocols:
            pickle.__init__(dumps, loads, proto)
            for obj in OBJS:
                self.do_pickle(obj, pickle)
            self.do_pickle(OBJS, pickle)

    @unittest.skipIf(marshal is None, "marshal")
    def testMarshal(self):
        pickle = self.pickle
        dumps = marshal.dumps
        loads = marshal.loads
        protocols = [0, 1, 2, 3, 4]
        protocols.append(None)
        for protocol in protocols:
            pickle.__init__(dumps, loads, protocol)
            for obj in OBJS:
                self.do_pickle(obj, pickle)
            self.do_pickle(OBJS, pickle)

    @unittest.skipIf(json is None, "json")
    def testJson(self):
        pickle = self.pickle

        def dumps(o):
            return json.dumps(o).encode()

        def loads(s):
            return json.loads(tobytes(s).decode())

        pickle.__init__(dumps, loads)
        OBJS2 = [o for o in OBJS if not isinstance(o, (float, complex, tuple))]
        for obj in OBJS2:
            self.do_pickle(obj, pickle)
        self.do_pickle(OBJS2, pickle)

    @unittest.skipIf(yaml is None, "yaml")
    def testYAML(self):
        pickle = self.pickle

        def dumps(o):
            return yaml.dump(o).encode()

        def loads(s):
            return yaml.load(tobytes(s).decode(), Loader=yaml.Loader)

        pickle.__init__(dumps, loads)
        OBJS2 = [o for o in OBJS if not isinstance(o, (complex, tuple))]
        for obj in OBJS2:
            self.do_pickle(obj, pickle)
        self.do_pickle(OBJS2, pickle)


if __name__ == "__main__":
    unittest.main()



cdef class Status:
    """
    Status object.
    """

    def __cinit__(self, Status status: Status | None = None):
        cdef MPI_Status *s = &self.ob_mpi
        CHKERR( MPI_Status_set_source (s, MPI_ANY_SOURCE ) )
        CHKERR( MPI_Status_set_tag    (s, MPI_ANY_TAG    ) )
        CHKERR( MPI_Status_set_error  (s, MPI_SUCCESS    ) )
        if status is None: return
        self.ob_mpi = status.ob_mpi

    def __richcmp__(self, other, int op):
        if not isinstance(other, Status): return NotImplemented
        cdef Status s = <Status>self, o = <Status>other
        cdef int ne = memcmp(&s.ob_mpi, &o.ob_mpi, sizeof(MPI_Status))
        if   op == Py_EQ: return (ne == 0)
        elif op == Py_NE: return (ne != 0)
        cdef str mod = type(self).__module__
        cdef str cls = type(self).__name__
        raise TypeError(f"unorderable type '{mod}.{cls}'")

    def __reduce__(self) -> tuple[Any, tuple[Any, ...], dict[str, Any]]:
        return (__newobj__, (type(self),), self.__getstate__())

    def __getstate__(self) -> dict[str, int]:
        cdef dict state = {
            'source': self.Get_source(),
            'tag': self.Get_tag(),
            'error': self.Get_error(),
        }
        try:
            state['count'] = self.Get_elements(__BYTE__)
        except NotImplementedError:  # ~> legacy
            pass                     # ~> legacy
        try:
            state['cancelled'] = self.Is_cancelled()
        except NotImplementedError:  # ~> legacy
            pass                     # ~> legacy
        return state

    def __setstate__(self, state: dict[str, int]) -> None:
        self.Set_source(state['source'])
        self.Set_tag(state['tag'])
        self.Set_error(state['error'])
        if 'count' in state:
            self.Set_elements(__BYTE__, state['count'])
        if 'cancelled' in state:
            self.Set_cancelled(state['cancelled'])

    def Get_source(self) -> int:
        """
        Get message source.
        """
        cdef int source = MPI_ANY_SOURCE
        CHKERR( MPI_Status_get_source(&self.ob_mpi, &source) )
        return source

    def Set_source(self, int source: int) -> None:
        """
        Set message source.
        """
        CHKERR( MPI_Status_set_source(&self.ob_mpi, source) )

    property source:
        """Message source."""
        def __get__(self) -> int:
            return self.Get_source()

        def __set__(self, value: int):
            self.Set_source(value)

    def Get_tag(self) -> int:
        """
        Get message tag.
        """
        cdef int tag = MPI_ANY_TAG
        CHKERR( MPI_Status_get_tag(&self.ob_mpi, &tag) )
        return tag

    def Set_tag(self, int tag: int) -> None:
        """
        Set message tag.
        """
        CHKERR( MPI_Status_set_tag(&self.ob_mpi, tag) )

    property tag:
        """Message tag."""
        def __get__(self) -> int:
            return self.Get_tag()

        def __set__(self, value: int):
            self.Set_tag(value)

    def Get_error(self) -> int:
        """
        Get message error.
        """
        cdef int error = MPI_SUCCESS
        CHKERR( MPI_Status_get_error(&self.ob_mpi, &error) )
        return error

    def Set_error(self, int error: int) -> None:
        """
        Set message error.
        """
        CHKERR( MPI_Status_set_error(&self.ob_mpi, error) )

    property error:
        """Message error."""
        def __get__(self) -> int:
            return self.Get_error()

        def __set__(self, value: int):
            self.Set_error(value)

    def Get_count(self, Datatype datatype: Datatype = BYTE) -> int:
        """
        Get the number of *top level* elements.
        """
        cdef MPI_Datatype dtype = datatype.ob_mpi
        cdef MPI_Count count = MPI_UNDEFINED
        CHKERR( MPI_Get_count_c(&self.ob_mpi, dtype, &count) )
        return count

    property count:
        """Byte count."""
        def __get__(self) -> int:
            return self.Get_count(__BYTE__)

        def __set__(self, value: int):
            self.Set_elements(__BYTE__, value)

    def Get_elements(self, Datatype datatype: Datatype) -> int:
        """
        Get the number of basic elements in a datatype.
        """
        cdef MPI_Datatype dtype = datatype.ob_mpi
        cdef MPI_Count elements = MPI_UNDEFINED
        CHKERR( MPI_Get_elements_c(&self.ob_mpi, dtype, &elements) )
        return elements

    def Set_elements(
        self,
        Datatype datatype: Datatype,
        Count count: int,
    ) -> None:
        """
        Set the number of elements in a status.

        .. note:: This method should be only used when implementing
           query callback functions for generalized requests.
        """
        cdef MPI_Datatype dtype = datatype.ob_mpi
        CHKERR( MPI_Status_set_elements_c(&self.ob_mpi, dtype, count) )

    def Is_cancelled(self) -> bool:
        """
        Test to see if a request was cancelled.
        """
        cdef int flag = 0
        CHKERR( MPI_Test_cancelled(&self.ob_mpi, &flag) )
        return <bint>flag

    def Set_cancelled(self, bint flag: bool) -> None:
        """
        Set the cancelled state associated with a status.

        .. note:: This method should be used only when implementing
           query callback functions for generalized requests.
        """
        CHKERR( MPI_Status_set_cancelled(&self.ob_mpi, flag) )

    property cancelled:
        """Cancelled state."""
        def __get__(self) -> bool:
            return self.Is_cancelled()

        def __set__(self, value: bool):
            self.Set_cancelled(value)

    # Memory
    # ------

    def tomemory(self) -> memoryview:
        """
        Return status memory view.
        """
        cdef buffer buf = newbuffer()
        cdef void *base = <void*> &self.ob_mpi
        cdef Py_ssize_t size = <Py_ssize_t> sizeof(MPI_Status)
        PyBuffer_FillInfo(&buf.view, self, base, size, 0, PyBUF_SIMPLE)
        return PyMemoryView_FromObject(buf).cast('i')

    # Fortran Handle
    # --------------

    def py2f(self) -> list[int]:
        """
        """
        cdef MPI_Fint f_status[16]
        cdef MPI_Status *c_status = &self.ob_mpi
        cdef Py_ssize_t size = MPI_F_STATUS_SIZE
        CHKERR( MPI_Status_c2f(c_status, f_status) )
        return [f_status[i] for i in range(size)]

    @classmethod
    def f2py(cls, arg: list[int]) -> Self:
        """
        """
        cdef MPI_Fint f_status[16]
        cdef MPI_Status c_status[1]
        cdef Py_ssize_t size = MPI_F_STATUS_SIZE
        for i in range(size): f_status[i] = arg[i]
        CHKERR( MPI_Status_f2c(f_status, c_status) )
        return PyMPIStatus_New(c_status)


F_SOURCE      = MPI_F_SOURCE
F_TAG         = MPI_F_TAG
F_ERROR       = MPI_F_ERROR
F_STATUS_SIZE = MPI_F_STATUS_SIZE



import argparse
import collections
import os
import shutil
import tempfile
import textwrap
from pathlib import Path

from wheel.cli import pack as wheel_pack
from wheel.wheelfile import WheelFile


def zip_extract(zfile, zinfo, destination):
    zfile.extract(zinfo, destination)
    # https://github.com/python/cpython/issues/59999
    permissions = zinfo.external_attr >> 16 & 0o777
    destination.joinpath(zinfo.filename).chmod(permissions)


parser = argparse.ArgumentParser()
parser.add_argument("wheelhouse", nargs="?", default="wheelhouse")
parser.add_argument("output_dir", nargs="?", default="dist")
opts = parser.parse_args()

wheelhouse = Path(opts.wheelhouse)
output_dir = Path(opts.output_dir)
working_dir = Path(tempfile.mkdtemp())
ext_suffix = ".so" if os.name == "posix" else ".pyd"
extensions = ["mpi4py.MPI"]

shutil.rmtree(output_dir, ignore_errors=True)
output_dir.mkdir(parents=True, exist_ok=True)

wheels = collections.defaultdict(list)
for whl in sorted(wheelhouse.glob("*.whl")):
    dist, version, py, abi, plat = whl.stem.split("-")
    if "+" in version:  # local version
        package = dist
        version, sep, variant = version.partition("+")
        variant = sep + variant
    else:  # dist suffix
        package, sep, variant = dist.partition("_")
        variant = sep + variant
    wheels[package, version, (py, abi, plat)].append(variant)

for (package, version, tags), variantlist in wheels.items():
    namever = f"{package}-{version}"
    wheeltags = "-".join(tags)
    root_dir = working_dir / namever
    package_dir = root_dir / package.partition("_")[0]
    distinfo_dir = root_dir / f"{namever}.dist-info"
    ext_suffix_glob = f".*{ext_suffix}"
    if tags[1] == "abi3" and tags[2].startswith("win"):
        ext_suffix_glob = f"{ext_suffix}"

    variant_registry = []
    for i, variant in enumerate(sorted(variantlist)):
        if variant[0] == "+":  # local version
            dist = f"{package}"
            distver = f"{dist}-{version}{variant}"
        if variant[0] == "_":  # dist suffix
            dist = f"{package}{variant}"
            distver = f"{dist}-{version}"

        wheelname = f"{distver}-{wheeltags}.whl"
        wheelpath = wheelhouse / wheelname

        if i == 0:
            with WheelFile(wheelpath) as wf:
                message = f"Unpacking wheel {wheelpath}..."
                print(message, end="", flush=True)
                for zinfo in wf.filelist:
                    zip_extract(wf, zinfo, root_dir)
                print("OK", flush=True)

            distinfo_dir.with_stem(distver).rename(distinfo_dir)

            for extension in extensions:
                extpath = Path().joinpath(*extension.split("."))
                for extfile in root_dir.glob(f"{extpath}{ext_suffix_glob}"):
                    extfile.unlink()

            for libdir in (
                Path(dist).with_suffix(".libs"),
                Path(package).with_suffix(".libs"),
                Path(package) / ".libs",
            ):
                libdir = root_dir / libdir
                if libdir.exists():
                    libdir.rmdir()

            mpipth = root_dir / "mpi.pth"
            if mpipth.exists():
                mpipth.unlink()

            record = distinfo_dir / "RECORD"
            record.unlink()

            metadata = distinfo_dir / "METADATA"
            data = metadata.read_text(encoding="utf-8")
            data = data.replace(variant.replace("_", "-"), "")
            data = data.replace(variant, "")
            metadata.write_text(data, encoding="utf-8")

        transtb = str.maketrans("_.", "--")
        variant = variant[1:].translate(transtb)
        variant_registry.append(variant)
        with WheelFile(wheelpath) as wf:
            extract = []
            for zinfo in wf.filelist:
                member = Path(zinfo.filename)
                for extension in extensions:
                    extpath = Path().joinpath(*extension.split("."))
                    if member.match(f"{extpath}{ext_suffix_glob}"):
                        extract.append(zinfo)
            for zinfo in extract:
                member = Path(zinfo.filename)
                message = f"Extracting: {member} [{variant}]..."
                print(message, end="", flush=True)
                zip_extract(wf, zinfo, root_dir)
                extension = root_dir.joinpath(member)
                extname, suffix = extension.name.split(".", 1)
                filename = f"{extname}.{variant}.{suffix}"
                extension.rename(extension.parent / filename)
                print("OK", flush=True)

    source = package_dir / "__init__.py"
    with source.open("a", encoding="utf-8") as fh:
        fh.write(
            textwrap.dedent("""\n
        # Install MPI ABI finder
        from . import _mpiabi  # noqa: E402
        _mpiabi._install_finder()
        """)
        )
        if variant_registry:
            fh.write("# Register MPI ABI variants\n")
        for variant in variant_registry:
            for module in extensions:
                fh.write(f"_mpiabi._register({module!r}, {variant!r})\n")
        if tags[2].startswith("win"):
            fh.write(
                textwrap.dedent("""\
            # Set Windows DLL search path
            __import__('_mpi_dll_path').install()
            """)
            )

    output_dir.mkdir(parents=True, exist_ok=True)
    wheel_pack.pack(root_dir, output_dir, None)
    shutil.rmtree(working_dir, ignore_errors=True)
    print(flush=True)



import os
import sys

import mpi4py

del mpi4py.rc
del sys.modules["mpi4py.rc"]

os.environ["MPI4PY_RC_ERRORS"] = "exception"

from mpi4py import MPI  # noqa: E402

assert MPI.Is_initialized()
assert not MPI.Is_finalized()
